from copy import deepcopy

import numpy as np
from numba import jit

import exosim.tasks.instrument as instrument
from exosim.tasks.task import Task
from exosim.utils.focal_plane_locations import locate_wavelength_windows
from exosim.utils.klass_factory import find_task
from exosim.utils.psf import create_psf


class PopulateFocalPlane(Task):
    """
    It populates the empty focal plane with monocromatic PSFs.

    Returns
    -------
    :class:`~exosim.models.signal.Signal`
        focal plane array populated
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        focal_plane: :class:`~exosim.models.signal.Signal`
            focal plane array (with time evolution)
        sources:  dict
            dictionary containing :class:`~exosim.models.signal.Sed`
        pointing: (:class:`astropy.units.Quantity`, :class:`astropy.units.Quantity`) (optional)
            telescope pointing direction, expressed ad a tuple of RA and DEC in degrees. Default is ``None``
        psf: :class:`numpy.ndarray` (optional)
            PSF array. Default is ``None``
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("focal_plane", "focal plane")
        self.add_task_param("sources", "sources")
        self.add_task_param("psf", "psf array", None)
        self.add_task_param("output", "output file", None)
        self.add_task_param("pointing", "telescope pointing", None)

    def execute(self):
        parameters = self.get_task_param("parameters")
        focal_plane = self.get_task_param("focal_plane")
        sources = self.get_task_param("sources")
        output = self.get_task_param("output")
        psf = self.get_task_param("psf")
        pointing = self.get_task_param("pointing")

        # load the psf
        psf = (
            self.load_psf(parameters, focal_plane, output)
            if psf is None
            else psf
        )

        i0_, j0_ = locate_wavelength_windows(psf, focal_plane, parameters)

        if sources:
            for source in sources.keys():
                self.info("populating focal plane with {}".format(source))

                j0 = deepcopy(j0_)
                i0 = deepcopy(i0_)

                # applying pointing offset in units of sub pixels
                compute_offset = instrument.ComputeSourcesPointingOffset()
                offset_spectral, offset_spatial = compute_offset(
                    source=sources[source].metadata,
                    pointing=pointing,
                    parameters=parameters,
                )

                j0 += offset_spectral
                i0 += offset_spatial

                focal_plane.data = populate(
                    i0, j0, psf, focal_plane.data, sources[source].data
                )

            focal_plane.data_units = sources[source].data_units
        # focal_plane.data.astype(np.float64)

        self.set_output([focal_plane, psf])

    @staticmethod
    def load_psf(parameters, focal_plane, output):
        """
        Loads the PSF as indicated in the configuration file or it creates them.

        Parameters
        ----------
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        focal_plane: :class:`~exosim.models.signal.Signal`
            focal plane array (with time evolution)
        output: :class:`~exosim.output.output.Output` (optional)
            output file

        Returns
        -------
        :class:`~numpy.ndarray`
            four dimensional array: axis 0 is time, axis 1 is wavelength,
            axis 2 is spatial, axis 3 is spectral.


        """
        if "psf_task" in parameters["psf"].keys():
            psf_task = find_task(
                parameters["psf"]["psf_task"], instrument.LoadPsf
            )
            psf_instance = psf_task()
            psf, _ = psf_instance(
                filename=parameters["psf"]["filename"],
                parameters=parameters,
                wavelength=focal_plane.spectral * focal_plane.spectral_units,
                time=focal_plane.time * focal_plane.time_units,
                output=output,
            )
        # create the psf
        else:
            nzero = (
                parameters["psf"]["nzero"]
                if "nzero" in parameters["psf"].keys()
                else 4
            )
            if "size" in parameters["psf"].keys():
                array_size = [
                    parameters["psf"]["size_y"],
                    parameters["psf"]["size_x"],
                ]
            else:
                array_size = None

            psf = create_psf(
                focal_plane.spectral * focal_plane.spectral_units,
                (parameters["Fnum_x"], parameters["Fnum_y"]),
                focal_plane.metadata["focal_plane_delta"],
                shape=parameters["psf"]["shape"].lower(),
                nzero=nzero,
                max_array_size=focal_plane.data[0].shape,
                array_size=array_size,
            )
            psf = psf[np.newaxis, ...]
            psf = np.repeat(psf, focal_plane.time.size, axis=0)

            # store the psf
            if output is not None:
                output_group = output.create_group("psf")
                output_group.write_array(
                    "psf_cube",
                    psf,
                    metadata={
                        "spatial fp_axis": 1,
                        "spectral_fp_axis": 2,
                        "wavelength_axis": 0,
                    },
                )
                norms = psf.sum(axis=0).sum(axis=0)
                output_group.write_array("norm", norms)
                output_group.write_quantity(
                    "wavelength",
                    focal_plane.spectral * focal_plane.spectral_units,
                )

        return psf


@jit(nopython=True)
def populate(
    i0: np.array,
    j0: np.array,
    psf: np.array,
    focal_plane: np.array,
    source: np.array,
) -> np.array:
    """it populates the focal plane adding the pfs"""
    for k in range(len(i0)):
        # this section is to avoid the psf to be out of the focal plane
        # i and j are the indexes of the psf on the focal plane
        # i_start and j_start are the indexes of the start cropped psf on the psf array
        # i_stop and j_stop are the indexes of the end cropped psf on the psf array

        # set the expected end of the psf
        i1 = i0[k] + psf.shape[2]
        j1 = j0[k] + psf.shape[3]

        # psf starting index
        i_start, j_start = 0, 0
        # check that teh expected start is not out of the focal plane
        if i0[k] < 0:
            i_start = -i0[k]
            i0[k] = 0
        if j0[k] < 0:
            j_start = -j0[k]
            j0[k] = 0

        # set the actual end of the psf
        i_stop, j_stop = i_start + psf.shape[2], j_start + psf.shape[3]

        # check that the expected end is not out of the focal plane
        if i1 > focal_plane.shape[1]:
            i_stop = i_start + focal_plane.shape[1] - i0[k]
            i1 = focal_plane.shape[1]
        if j1 > focal_plane.shape[2]:
            j_stop = j_start + focal_plane.shape[2] - j0[k]
            j1 = focal_plane.shape[2]

        for i in range(focal_plane.shape[0]):
            psfi = i
            if psf.shape[0] == 1:
                psfi = 0
            focal_plane[i, i0[k] : i1, j0[k] : j1] += (
                psf[psfi, k, i_start:i_stop, j_start:j_stop] * source[i, 0, k]
            )
    return focal_plane



import numpy as np
from astropy import units as u
from astropy.modeling.physical_models import BlackBody

import exosim.models.signal as signal
from exosim.tasks.task import Task


class CreateCustomSource(Task):
    """
    Creates a custom SED from input parameters.

    Returns
    -------
    :class:`~exosim.models.signal.Sed`
        Star Sed

    Raises
    ------
    TypeError:
        if the output is not a :class:`~exosim.models.signal.Sed` class

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        -----------
        parameters: dict
            dictionary containing the source parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """
        self.add_task_param("parameters", "source parameters")

    def execute(self):
        self.info("Create custom sed")
        parameters = self.get_task_param("parameters")

        sed = self.model(parameters)

        # checking output
        if not isinstance(sed, signal.Sed):
            self.error("wrong output format")
            raise TypeError("wrong output format")

        self.debug("sed: {}".format(sed.data))

        self.set_output(sed)

    def model(self, parameters):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the source parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`.

        Returns
        --------
        :class:`~exosim.models.signal.Sed`
            source sed

        """
        R = parameters["R"].si
        D = parameters["D"].si
        T = parameters["T"].si

        wl_min = parameters["wl_min"]
        wl_max = parameters["wl_max"]
        n_points = parameters["n_points"]
        wl = np.linspace(wl_min, wl_max, n_points)

        omega_star = np.pi * (R / D) ** 2 * u.sr
        bb = BlackBody(T)
        bb_ = bb(wl).to(u.W / u.m**2 / u.sr / u.um, u.spectral_density(wl))
        sed = signal.Sed(spectral=wl, data=omega_star * bb_)

        return sed



from exosim.tasks.task import Task


class Multiaccum(Task):
    """
    It computes the MULTIACCUM gain factor from Rauscher and Fox 2007 (http://iopscience.iop.org/article/10.1086/520887/pdf)
    with the correction from Robberto 2009, also reported in Batalha 2017 (https://doi.org/10.1088/1538-3873/aa65b0)

    Returns
    ---------
    float:
        read noise gain factor
    float:
        shot noise gain factor
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the MULTIACCUM channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """
        self.add_task_param("parameters", "multiaccum parameters dict")

    def execute(self):
        self.debug("computing MULTIACCUM")
        parameters = self.get_task_param("parameters")

        n = parameters["n"]
        m = parameters["m"]
        tf = parameters["tf"]
        tg = parameters["tg"]

        if n < 2:
            n = 2.0  # Force to CDS in nRead < 2

        read_gain = 12.0 * (n - 1.0) / (m * (n**2 + n))
        shot_gain = (
            6.0 / 5.0 * (n**2 + 1.0) / (n**2 + n) * (n - 1.0) * tg
            + (m**2 - 1.0) * (n - 1.0) / (m * (n**2 + n)) * tf
        )

        self.set_output([read_gain, shot_gain])



# This code is inspired by the code developed for TauREx 3.1.
# Therefore, we attach here TauREx3.1 license:
#
# BSD 3-Clause License
#
# Copyright (c) 2019, Ahmed F. Al-Refaie, Quentin Changeat, Ingo Waldmann, Giovanna Tinetti
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import datetime
import os
from typing import List, Union

import astropy.units as u
import h5py
import numpy as np
from astropy.table import QTable, Table, meta

import exosim.output.output as output
from exosim import (
    __author__,
    __branch__,
    __citation__,
    __commit__,
    __copyright__,
    __license__,
    __pkg_name__,
    __title__,
    __url__,
    __version__,
)
from exosim.utils.runConfig import RunConfig

META_KEY = "__table_column_meta__"


class HDF5OutputGroup(output.OutputGroup):
    def __init__(
        self,
        entry: h5py.Group,
        fd: h5py.File,
        cache: bool = False,
        filename: str = None,
    ) -> None:
        self.set_log_name()
        self.fd = fd
        self._entry = entry
        self._cache = cache
        self.filename = filename

    def write_array(
        self, array_name: str, array: np.ndarray, metadata: dict = None
    ) -> None:
        if isinstance(array, list):
            for idx, a in enumerate(array):
                self.write_array("{}{}".format(array_name, idx), a, metadata)
            return
        ds = self._entry.create_dataset(
            str(array_name), data=array, shape=array.shape, dtype=array.dtype
        )

        ds.attrs["datatype"] = str(array.__class__)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_table(
        self,
        table_name: str,
        table: Union[Table, QTable],
        metadata: dict = None,
        replace: bool = False,
    ) -> None:
        if replace:
            if str(table_name) in self._entry.keys():
                del self._entry[str(table_name)]
                try:
                    del self._entry[str(table_name) + "." + META_KEY]
                except KeyError:
                    pass
                try:
                    del self._entry["{}_to_group".format(str(table_name))]
                except KeyError:
                    pass

        table = _encode_mixins(table)
        if any(col.info.dtype.kind == "U" for col in table.itercols()):
            table = table.copy(copy_data=False)
            table.convert_unicode_to_bytestring()

        self._entry.create_dataset(str(table_name), data=table.as_array())
        header_yaml = meta.get_yaml_from_table(table)

        header_encoded = [h.encode("utf-8") for h in header_yaml]
        self._entry.create_dataset(
            str(table_name) + "." + META_KEY, data=header_encoded
        )

        tg = self._entry.create_group("{}_to_group".format(str(table_name)))
        for col in table.keys():
            tg_c = tg.create_group(str(col))
            tg_c.create_dataset("value", data=table[col])
            if table[col].unit is not None:
                tg_c.create_dataset("unit", data=str(table[col].unit))

        tg.attrs["datatype"] = str(table.__class__)
        if metadata:
            for k, v in metadata.items():
                tg.attrs[k] = v
            for k, v in table.meta.items():
                tg.attrs[k] = v

    def write_scalar(
        self, scalar_name: str, scalar: float, metadata: dict = None
    ) -> None:
        ds = self._entry.create_dataset(str(scalar_name), data=scalar)

        ds.attrs["datatype"] = str(scalar.__class__)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_string(
        self, string_name: str, string: str, metadata: dict = None
    ) -> None:
        # h5py new string hanlder support bites. So we convert strings into bites
        string = string.encode("utf-8")
        ds = self._entry.create_dataset(str(string_name), data=string)

        ds.attrs["datatype"] = str(string.__class__)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def create_group(self, group_name: str) -> "HDF5OutputGroup":
        """
        it creates an :class:`~exosim.output.hdf5.hdf5.HDF5OutputGroup`.

        Parameters
        ----------
        group_name: str
            group name

        Returns
        -------
        OutputGroup

        """
        entry = None
        if self._entry:
            try:
                entry = self._entry.create_group(str(group_name))
            except ValueError:
                entry = self._entry[str(group_name)]
        return HDF5OutputGroup(
            entry, fd=self.fd, cache=self._cache, filename=self.filename
        )

    def delete_data(self, key):
        try:
            del self._entry[key]
        except KeyError:
            self.error("key {} not found".format(key))
            raise KeyError

    def flush(self) -> None:
        self.fd.flush()

    def write_string_array(
        self, string_name: str, string_array: List[str], metadata: dict = None
    ) -> None:
        string_array = string_array.astype("S64")
        ds = self._entry.create_dataset(str(string_name), data=string_array)

        ds.attrs["datatype"] = str(string_array.__class__)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_quantity(
        self, quantity_name: str, quantity: u.Quantity, metadata: dict = None
    ) -> None:
        if quantity_name == "value":
            qg_c = self._entry.create_dataset(
                "value",
                data=quantity.value,
            )
        else:
            qg_c = self._entry.create_group(str(quantity_name))
            qg_c.create_dataset(
                "value",
                data=quantity.value,
            )
            qg_c.create_dataset("unit", data=str(quantity.unit))
            pass
        qg_c.attrs["datatype"] = str(quantity.__class__)

        if metadata:
            for k, v in metadata.items():
                qg_c.attrs[k] = v


def _encode_mixins(tbl):
    from astropy.table import serialize
    from astropy.utils.data_info import serialize_context_as

    with serialize_context_as("hdf5"):
        encode_tbl = serialize.represent_mixins_as_columns(tbl)

    return encode_tbl


class HDF5Output(output.Output):
    def __init__(self, filename, append=False, cache=False):
        super().__init__()
        self.filename = filename
        self._append = append
        self._cache = cache
        self.fd = None

    def open(self) -> None:
        self.fd = self._openFile(self.filename)
        self.debug("opened {}".format(self.filename))

    def _openFile(self, fname: str) -> h5py.File:
        mode = "w"
        if self._append:
            mode = "a"

        rdcc_w0 = None
        if self._cache:
            rdcc_w0 = 1

        attrs = {
            "file_name": fname,
            "file_time": datetime.datetime.now().isoformat(),
            "creator": self.__class__.__name__,
            "HDF5_Version": h5py.version.hdf5_version,
            "h5py_version": h5py.version.version,
            "program_name": str(__title__),
            "package name": str(__pkg_name__),
            "program_version": str(__version__),
            "author": str(__author__),
            "copyright": str(__copyright__),
            "license": str(__license__),
            "url": str(__url__),
            "citation": str(__citation__),
            "git commit": str(__commit__),
            "git branch": str(__branch__),
        }

        fd = h5py.File(fname, mode=mode, rdcc_w0=rdcc_w0)
        for key in attrs:
            fd.attrs[key] = attrs[key]

        if mode == "w" or "info" not in fd.keys():
            try:
                gd_ = fd["info"]
            except KeyError:
                gd_ = fd.create_group("info")
            gd = HDF5OutputGroup(gd_, fd)
            gd.store_dictionary(attrs, "exosim")
            gd.store_dictionary(RunConfig.__dict__(), "RunConfig")

        fd.flush()

        return fd

    def add_info(self, attrs: dict, name: str = None) -> None:
        gd = self.create_group("info")
        gd.store_dictionary(attrs, name)

    def flush(self) -> None:
        self.fd.flush()

    def create_group(self, group_name: str) -> "HDF5OutputGroup":
        """
        it creates an :class:`~exosim.output.hdf5.hdf5.HDF5OutputGroup`.

        Parameters
        ----------
        group_name: str
            group name

        Returns
        -------
        OutputGroup

        """

        entry = None
        if self.fd:
            try:
                entry = self.fd.create_group(str(group_name))
            except ValueError:
                entry = self.fd[str(group_name)]
        return HDF5OutputGroup(
            entry, cache=self._cache, filename=self.filename, fd=self.fd
        )

    def store_dictionary(
        self, dictionary: dict, group_name: str = None
    ) -> None:
        """
        it stores a full dictionary inside the :class:`HDF5Output`, forcing the flush.

        Parameters
        ----------
        dictionary: dict
            data to store
        group_name: str (optional)
            group name for the data
        """

        from exosim.output.utils import (
            recursively_save_dict_contents_to_output,
        )

        out = self
        if group_name is not None:
            out = self.create_group(group_name)

        recursively_save_dict_contents_to_output(out, dictionary)

        self.fd.flush()

    def close(self) -> None:
        if self.fd:
            self.fd.flush()
            self.fd.close()
            self.debug("closed {}".format(self.filename))

    def getsize(self) -> u.Quantity:
        """It returns the output file size"""
        size = os.path.getsize(self.filename) * u.b
        if size.to(u.Tb).value > 1.0:
            size = size.to(u.Tb)
        elif size.to(u.Gb).value > 1.0:
            size = size.to(u.Gb)
        elif size.to(u.Mb).value > 1.0:
            size = size.to(u.Mb)
        return size

    def delete_data(self, key):
        try:
            del self.fd[key]
        except KeyError:
            self.error("key {} not found".format(key))
            raise KeyError



import importlib.metadata as metadata
import os.path
from datetime import date

__version__ = metadata.version("exosim")

meta = metadata.metadata("exosim")
__pkg_name__ = meta.get("Name", "exosim")
__title__ = "ExoSim2"
__url__ = meta.get(
    "Home-page", "https://github.com/arielmission-space/ExoSim2-public"
)
__author__ = meta.get("Author", "L. V. Mugnai")
__license__ = meta.get("License", "BSD-3-Clause")
__copyright__ = f"2020-{date.today().year}, {__author__}"
__citation__ = """@article{Mugnai2025,
    author  = {Mugnai, Lorenzo V. and Bocchieri, Andrea and Pascale, Enzo and Lorenzani, Andrea and Papageorgiou, Andreas},
    title   = {ExoSim 2: the new exoplanet observation simulator applied to the Ariel space mission},
    journal = {Experimental Astronomy},
    year    = {2025},
    volume  = {59},
    number  = {1},
    pages   = {9},
    doi     = {10.1007/s10686-024-09976-2},
    url     = {https://doi.org/10.1007/s10686-024-09976-2},
    eprint        = {2501.12809},
    archivePrefix = {arXiv},
    primaryClass  = {astro-ph.IM},
    ascl_id       = {ascl:2503.031},
    }"""
# load package commit number
try:
    __base_dir__ = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
except NameError:
    __base_dir__ = None

__commit__ = None
__branch__ = None
if __base_dir__ is not None and os.path.exists(
    os.path.join(__base_dir__, ".git")
):
    git_folder = os.path.join(__base_dir__, ".git")
    with open(os.path.join(git_folder, "HEAD")) as fp:
        ref = fp.read().strip()
    ref_dir = ref[5:]
    __branch__ = ref[16:]
    try:
        with open(os.path.join(git_folder, ref_dir)) as fp:
            __commit__ = fp.read().strip()
    except FileNotFoundError:
        __commit__ = None


# initialise logger
import logging

logger = logging.getLogger(__pkg_name__)
logger.info("code version {}".format(__version__))

from exosim.log import setLogLevel

setLogLevel(logging.INFO)



import gc
import os
import re
from copy import deepcopy

import astropy.units as u
import h5py
import numpy as np
from scipy.interpolate import RectBivariateSpline, interp1d

import exosim.tasks.instrument.loadPsf as loadPsf
from exosim.utils.aperture import find_rectangular_aperture


class LoadPsfPaos(loadPsf.LoadPsf):
    """
    It loads the PSFs from a PAOS file.
    Each PAOS file contains a PSF vs wavelength.
    The task loads the PSF cube provided by the `filename` parameter.
    The PSF are then interpolated over a grid matching the one used to produce the focal planes,
    to convert them into the physical units.
    Then the total volume of the interpolated PSF is rescaled to the total volume of the original one.
    This allow to take into account for loss in the transmission due to the optical path.
    The PSF are then interpolated over a wavelength grid matching the one used to for the focal plane, producing the cube.
    Then, the PSF cube is repeated on the temporal axis, because no temporal variation is considered in this Task.

    Returns
    -------
    :class:`~numpy.ndarray`
        cube of psfs.  axis=0 is time, axis=1 is wavelength, axis=2 is spatial direction, axis=3 is spectral direction.
    """

    def model(self, filename, parameters, wavelength, time):
        self.debug("loading PAOS psf")

        if (
            "psf" in parameters
            and "time_dependence" in parameters["psf"].keys()
        ):
            timeDependence = parameters["psf"]["time_dependence"]
        else:
            timeDependence = True
        if "oversampling" in parameters["detector"].keys():
            oversampling = parameters["detector"]["oversampling"]
        else:
            oversampling = 1
        delta_pix = parameters["detector"]["delta_pix"] / oversampling
        with h5py.File(os.path.expanduser(filename), "r") as data:
            # find the sampled wl
            temp = re.compile(r"\d+(?:\.\d*)")
            wl_sampled_h5 = [ele for ele in data.keys() if temp.match(ele)]

            # load psf interpolated into the window
            psf_cube = [
                self.load_imac(
                    wl,
                    data,
                    parameters,
                    oversampling,
                    delta_pix,
                )
                for wl in wl_sampled_h5
            ]
            wl_sampled = [
                self.load_wl_sampled(wl, data) for wl in wl_sampled_h5
            ]

        wl_sampled = np.array(wl_sampled) * u.m
        psf_cube = np.array(psf_cube)

        # wl_interpolate to the output wl grid
        psf_out = self.wl_interpolate(
            psf_cube, wavelength, wl_sampled.to(u.um)
        )

        del psf_cube
        gc.collect()

        # repeat over time
        psf_cube_out = psf_out[np.newaxis, ...]
        del psf_out
        gc.collect()

        if timeDependence:
            psf_cube_out = np.repeat(psf_cube_out, time.size, axis=0)
        return psf_cube_out

    @staticmethod
    def load_wl_sampled(wl, data):
        """
        extract the wavelength from the data stored in the surfaces

        Parameters
        ----------
        wl: str
            wavelength
        data: :class:`h5py.File`
            opened HDF5 file

        Returns
        -------
        float:
            wavelength

        """
        data_wl_surfaces = list(data[wl].keys())
        data_wl_surfaces.sort()
        data_cube = data[wl][data_wl_surfaces[-1]]
        return float(data_cube["wl"][()])

    def load_imac(self, wl, data, parameters, oversampling, delta_pix):
        """
        Returns the PSF interpolated to the detector pixel size.
        The PSF is normalized to the initial volume after the interpolation.

        Parameters
        ----------
        wl: str
            wavelength
        data: :class:`h5py.File`
            opened HDF5 file
        parameters: dict
            detector description
        oversampling: int
            oversampling factor
        delta_pix: float or :class:`astropy.units.Quantity`
            sub-pixel size

        Returns
        -------
        :class:`numpy.ndarray`
            PSF scaled to physical size
        """

        data_wl_surfaces = list(data[wl].keys())
        data_wl_surfaces.sort()
        data_cube = data[wl][data_wl_surfaces[-1]]

        # load data
        dx = data_cube["dy"][()] * u.m
        dy = data_cube["dx"][()] * u.m
        ima = data_cube["amplitude"][()] * data_cube["amplitude"][()]
        # ima = ima.T
        vol = ima.sum()

        # scale the psf image
        paos_x, paos_y = np.arange(0, ima.shape[0]) * dx.to(u.um), np.arange(
            0, ima.shape[1]
        ) * dy.to(u.um)

        size = min(
            parameters["detector"]["spatial_pix"],
            parameters["detector"]["spectral_pix"],
        )
        spatial_size = size * oversampling
        spectral_size = size * oversampling

        self.debug("windows sizes: {}, {}".format(spatial_size, spectral_size))

        exosim_x = np.arange(0, spatial_size - 1) * delta_pix.to(u.um)
        exosim_y = np.arange(0, spectral_size - 1) * delta_pix.to(u.um)

        exosim_x -= (
            exosim_x[int(spatial_size // 2)] - paos_x[ima.shape[0] // 2]
        )
        exosim_y -= (
            exosim_y[int(spectral_size // 2)] - paos_y[ima.shape[1] // 2]
        )

        f = RectBivariateSpline(paos_x, paos_y, ima, kx=1, ky=1)
        imac = f(exosim_x.value, exosim_y.value)

        imac -= imac.min()

        # normalise
        norm = imac.sum()
        imac /= norm
        imac *= vol

        return np.array(imac)

    def crop_image_stack(self, psf_out, ene=0.99, time_iteration=False):
        """
        It crops the image stack.
        It takes the PSF for the longest wavelength, as it is expected to be the largest PSF,
        and then selects the smallest aperture which collect at least the desired ene,
        using :func:`~exosim.utils.psf.find_rectangular_aperture`.
        Then it remove these area from all the image in the psf data cube.

        Parameters
        ----------
        psf_out: :class:`numpy.ndarray`
            output PSF cube
        ene: float
            encircled energy desired. Default is 99%.
        time_iteration: bool
            if time iteration is True then it perform

        Returns
        -------
        :class:`numpy.ndarray`
            PSF cube cropped
        """

        h, w = 0, 0
        if time_iteration or psf_out.ndims > 3:
            self.debug("looking for best aperture: iterating over time")
            for t in range(psf_out.shape[0]):
                sizes, surf, ene = find_rectangular_aperture(
                    psf_out[t, -1, :, :], ene
                )
                h = sizes[0] if sizes[0] > h else h
                w = sizes[1] if sizes[1] > w else w

            crop_h = (psf_out.shape[2] - h) // 2
            crop_w = (psf_out.shape[3] - w) // 2

            return psf_out[
                :,
                :,
                int(crop_h) : int(crop_h + h),
                int(crop_w) : int(crop_w + w),
            ]

        else:
            (h, w), surf, ene = find_rectangular_aperture(
                psf_out[-1, :, :], ene
            )
            crop_h = (psf_out.shape[1] - h) // 2
            crop_w = (psf_out.shape[2] - w) // 2

            return psf_out[
                :, int(crop_h) : int(crop_h + h), int(crop_w) : int(crop_w + w)
            ]

    @staticmethod
    # @jit(nopython=True, parallel=True)
    def wl_interpolate(psf_cube, wl, wl_sampled):
        """
        This function interpolates the PSF to the desired wavelength grid

        Parameters
        ----------
        psf_out: :class:`numpy.ndarray`
            output PSF cube
        psf_cube: :class:`numpy.ndarray`
            input PSF cube
        wl: :class:`numpy.ndarray`
            desired wavelength grid
        wl_sampled: :class:`numpy.ndarray`
            input wavelength grid

        Returns
        -------
        :class:`numpy.ndarray`
            output PSF cube
        """
        # take into account different ordering
        if wl_sampled.size > 1:
            wl_unsorted = deepcopy(wl)
            wl.sort()

            psf_cube_resh = psf_cube.reshape(
                psf_cube.shape[0], psf_cube.shape[1] * psf_cube.shape[2]
            )
            f = interp1d(
                wl_sampled,
                psf_cube_resh,
                axis=0,
                bounds_error=False,
                fill_value="extrapolate",
            )
            psf_out = f(wl)
            psf_out = psf_out.reshape(
                wl.size, psf_cube.shape[1], psf_cube.shape[2]
            )
            idx = [int(np.where(wl == wl_)[0][0]) for wl_ in wl_unsorted]
            psf_out = psf_out[idx, ...]

        else:
            psf_out = np.repeat(psf_cube, wl.size, axis=0)

        # psf_out[k, ...] /= psf_out[k, ...].sum()
        return psf_out



from collections import OrderedDict
from typing import Tuple

import astropy.units as u
import numpy as np

import exosim.log as log
import exosim.models.signal as signal
import exosim.tasks.instrument as instrument
import exosim.tasks.parse as parse
from exosim.utils.klass_factory import find_and_run_task
from exosim.utils.types import ArrayType, OutputType, ValueType


class Channel(log.Logger):
    """
    It handles the channel gven the description

    Attributes
    ----------
    ch_name: str
        channel name
    path: dict
        dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
        represeting the radiance and efficiency of the path.
    responsivity: :class:`~exosim.models.signal.Signal`
        channel responsivity
    sources: dict
        dictionary containing :class:`~exosim.models.signal.Signal`
    time: :class:`~astropy.units.Quantity`
        time grid.
    parameters: dict
        dictionary contained the optical element parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
    wavelength: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
        wavelength grid. If no units are attached is considered as expressed in `um`.
    focal_plane: :class:`~exosim.models.signal.Signal`
        source focal plane
    frg_focal_plane: :class:`~exosim.models.signal.Signal`
        foreground focal plane
    frg_sub_focal_planes: dict
        dictionary of :class:`~exosim.models.signal.Signal`. It contains the sub focal planes produced by the radiances.
        This dictionary is produced only if at least one optical surface has ``isolate=True``.
        The sum of the sub focal planes returns the frg_focal_plane. If not surface has ``isolate=True``, the dictionary is empty.
    output: :class:`~exosim.output.output.Output`
        output file
    target_source: str
        name of the target source
    """

    def __init__(
        self,
        parameters: dict,
        wavelength: ArrayType,
        time: ArrayType,
        output: OutputType = None,
    ) -> None:
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the optical element parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
            wavelength grid. If no units are attached is considered as expressed in `um`.
        time: :class:`~astropy.units.Quantity`
            time grid.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        self.set_log_name()

        self.parameters = parameters
        self.wavelength = wavelength
        self.time = time
        self.ch_name = parameters["value"]

        self.output = output.create_group(self.ch_name) if output else None

        # init to None
        self.path, self.responsivity, self.sources, self.psf = (
            None,
            None,
            None,
            None,
        )
        (
            self.focal_plane,
            self.bkg_focal_plane,
            self.frg_focal_plane,
            self.frg_sub_focal_planes,
        ) = (
            None,
            None,
            None,
            None,
        )

    def parse_path(self, light_path: OrderedDict) -> dict:
        """
        It applies :class:`~exosim.tasks.parse.parsePath.ParsePath`

        Parameters
        ----------
        light_path: `~collections.OrderedDict` (optional)
            dictionary of contributes

        Returns
        -------
        dict
            dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
            represeting the radiance and efficiency of the path.

        Note
        ----
        The resulting information is also stored in the class under `path` attribute.
        """

        parsePath = parse.ParsePath()
        self.path = parsePath(
            parameters=self.parameters["optical_path"],
            wavelength=self.wavelength,
            time=self.time,
            output=self.output,
            light_path=light_path,
            group_name="path",
        )
        self.path["efficiency"].write(output=self.output, name="efficiency")
        return self.path

    def estimate_responsivity(self) -> signal.Signal:
        """
        It estimates the responsivity using the indicated :class:`~exosim.tasks.instrument.loadResponsivity.LoadResponsivity`

        Returns
        -------
        :class:`~exosim.models.signal.Signal`
            channel responsivity

        Note
        ----
        The resulting information is also stored in the class under `responsivity` attribute.
        """
        responsivity_instance = find_and_run_task(
            self.parameters["qe"],
            "responsivity_task",
            instrument.LoadResponsivity,
        )

        self.responsivity = responsivity_instance(
            parameters=self.parameters,
            wavelength=self.wavelength,
            time=self.time,
        )
        self.responsivity.write(output=self.output, name="responsivity")
        return self.responsivity

    def propagate_foreground(self) -> dict:
        """
        It multiplies each radiance in the path by the solid angle.

        Returns
        -------
        dict
            dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
            represeting the radiance and efficiency of the path.

        Note
        ----
        it updates the `path` attribute of this class
        """

        propagateForegrounds = instrument.PropagateForegrounds()
        self.path = propagateForegrounds(
            light_path=self.path,
            parameters=self.parameters,
            responsivity=self.responsivity,
        )
        return self.path

    def propagate_sources(self, sources: OrderedDict, Atel: ValueType) -> dict:
        """
        It propagates the sources though the channel,
        by applying :class:`~exosim.tasks.instrument.propagateSources.PropagateSources`

        Parameters
        __________
        sources:  dict
            dictionary containing :class:`~exosim.models.signal.Sed`
        Atel:  :class:`~astropy.units.Quantity`
            effective telescope Area

        Returns
        -------
        dict
            dictionary containing :class:`~exosim.models.signal.Signal`
        """

        out_sources = {}

        for source in sources.keys():
            out_sources[source] = signal.Sed(
                data=sources[source].data,
                spectral=sources[source].spectral,
                time=sources[source].time,
                metadata=sources[source].metadata,
            )

        propagateSources = instrument.PropagateSources()
        self.sources = propagateSources(
            sources=out_sources,
            Atel=Atel,
            efficiency=self.path["efficiency"],
            responsivity=self.responsivity,
        )

        return self.sources

    def create_focal_planes(self) -> signal.Signal:
        """
        It produces the empty focal planes

        Returns
        -------
         :class:`~exosim.models.signal.Signal`
            focal plane array (with time evolution)
        """

        createFocalPlane = instrument.CreateFocalPlane()
        focal_plane = createFocalPlane(
            parameters=self.parameters,
            efficiency=self.path["efficiency"],
            time=self.time,
            output=self.output,
            group_name="focal_plane",
        )
        self.focal_plane = focal_plane
        self.frg_focal_plane = focal_plane.copy(dataset_name="frg_focal_plane")
        if len(self.sources) > 1:
            self.bkg_focal_plane = focal_plane.copy(
                dataset_name="bkg_focal_plane"
            )
        return focal_plane

    def rescale_contributions(self) -> None:
        """
        It updated the contributions (sources and path)
        by rebinning them to the wavelength solution grid
        and multipling them by the wl solution gradient
        """

        def wl_gradient(x_wav_osr):
            d_x_wav_osr = np.zeros_like(x_wav_osr)
            idx = np.where(x_wav_osr > 0.0)
            d_x_wav_osr[idx] = np.gradient(x_wav_osr[idx])
            if np.any(d_x_wav_osr < 0):
                d_x_wav_osr *= -1.0
            return d_x_wav_osr

        d_spectral_wl = (
            wl_gradient(self.focal_plane.spectral)
            * self.focal_plane.spectral_units
        )

        # multiply sources by gradient
        if self.sources:
            for source in self.sources.keys():
                self.sources[source].spectral_rebin(self.focal_plane.spectral)
                self.sources[source] *= d_spectral_wl

        # multiply radiances by gradient
        if self.path:
            for rad in [k for k in self.path.keys() if "radiance" in k]:
                self.path[rad].spectral_rebin(self.focal_plane.spectral)
                self.path[rad] *= d_spectral_wl

    @property
    def target_source(self):
        # TODO test this
        if len(self.sources) == 1:
            return list(self.sources.keys())[0]

        target = [
            source
            for source, param in self.sources.items()
            if "source_target" in param.metadata["parsed_parameters"].keys()
            and param.metadata["parsed_parameters"]["source_target"] == True
        ]
        if len(target) > 1:
            self.error(
                "More than one target source found. Please check your input file."
            )
        self.debug(f"Target source is {target[0]}")
        return target[0]

    def populate_focal_plane(
        self, pointing: Tuple[u.Quantity, u.Quantity] = None
    ) -> signal.Signal:
        """
        It populates the empty focal plane with monocromatic PSFs.

        Parameters
        -----------
        pointing: (:class:`astropy.units.Quantity`, :class:`astropy.units.Quantity`) (optional)
            telescope pointing direction, expressed ad a tuple of RA and DEC in degrees. Default is ``None``

        Returns
        -------
        :class:`~exosim.models.signal.Signal`
            focal plane array populated
        """

        # selecting the target source
        target = self.target_source

        # storing binned target source
        sources_out = self.output.create_group("sources")
        self.sources[target].write(sources_out, name=target)

        # populates the focal plane
        populateFocalPlane = instrument.PopulateFocalPlane()
        focal_plane, psf = populateFocalPlane(
            parameters=self.parameters,
            focal_plane=self.focal_plane,
            sources={target: self.sources[target]},
            pointing=pointing,
            output=self.output,
        )
        self.psf = psf
        self.focal_plane = focal_plane
        return focal_plane

    def populate_bkg_focal_plane(
        self, pointing: Tuple[u.Quantity, u.Quantity] = None
    ) -> signal.Signal:
        """
        It populates the empty background focal plane with monocromatic PSFs for each of the background sources.

        Parameters
        -----------
        pointing: (:class:`astropy.units.Quantity`, :class:`astropy.units.Quantity`) (optional)
            telescope pointing direction, expressed ad a tuple of RA and DEC in degrees. Default is ``None``

        Returns
        -------
        :class:`~exosim.models.signal.Signal`
            background focal plane array populated
        """
        if len(self.sources) > 1:
            # TODO test this
            # removing target source from source dictionary
            target = self.target_source
            self.sources.pop(target)

            # storing binned sources
            sources_out = self.output.create_group("sources")
            for source in self.sources.keys():
                self.sources[source].write(sources_out, name=source)

            # populates the focal plane
            populateFocalPlane = instrument.PopulateFocalPlane()
            bkg_focal_plane, self.psf = populateFocalPlane(
                parameters=self.parameters,
                focal_plane=self.bkg_focal_plane,
                sources=self.sources,
                pointing=pointing,
                output=self.output,
                psf=self.psf,
            )
            self.bkg_focal_plane = bkg_focal_plane

        return self.bkg_focal_plane

    def apply_irf(self) -> signal.Signal:
        """
        It applies the intra pixel response function (IRF) to the focal plane


        Returns
        -------
        :class:`~exosim.models.signal.Signal`
            focal plane array
        """

        irf_instance = find_and_run_task(
            self.parameters["detector"],
            "irf_task",
            instrument.CreateIntrapixelResponseFunction,
        )

        self.parameters["psf_shape"] = self.psf.shape[1:]
        kernel, delta_kernel = irf_instance(
            parameters=self.parameters, output=self.output
        )

        if "convolution_method" in self.parameters["detector"]:
            convolution_method = self.parameters["detector"][
                "convolution_method"
            ]
        else:
            convolution_method = "fftconvolve"

        applyIntraPixelResponseFunction = find_and_run_task(
            self.parameters["detector"],
            "apply_irf_task",
            instrument.ApplyIntraPixelResponseFunction,
        )

        # apply IRF to target focal plane
        focal_plane = applyIntraPixelResponseFunction(
            focal_plane=self.focal_plane,
            irf_kernel=kernel,
            irf_kernel_delta=delta_kernel,
            convolution_method=convolution_method,
        )

        if self.bkg_focal_plane is not None:
            # apply IRF to background focal plane
            bkg_focal_plane = applyIntraPixelResponseFunction(
                focal_plane=self.bkg_focal_plane,
                irf_kernel=kernel,
                irf_kernel_delta=delta_kernel,
                convolution_method=convolution_method,
            )
        else:
            bkg_focal_plane = None

        return focal_plane, bkg_focal_plane

    def populate_foreground_focal_plane(
        self,
    ) -> Tuple[signal.Signal, signal.Signal]:
        """
        It adds the foreground contribution to the foreground focal plane

        Returns
        -------
        :class:`~exosim.models.signal.Signal`
            focal plane array
        """

        # populates the focal plane
        foregroundToFocalPlane = instrument.ForegroundsToFocalPlane()
        frg_focal_plane, sub_focal_planes = foregroundToFocalPlane(
            parameters=self.parameters,
            focal_plane=self.frg_focal_plane,
            path=self.path,
        )

        self.frg_focal_plane = frg_focal_plane
        self.frg_sub_focal_planes = sub_focal_planes
        return frg_focal_plane, sub_focal_planes



import logging
import os

import click
from rich_click import RichGroup

import exosim.recipes as recipes
from exosim import __version__
from exosim.log import addLogFile, setLogLevel
from exosim.utils import RunConfig

# Logger configuration
logger = logging.getLogger("exosim")
code_name_and_version = "ExoSim {}".format(__version__)


# Common options decorator for shared command-line arguments
def common_options(func):
    """Decorator to add common options to commands."""
    options = [
        click.option(
            "-c",
            "--configuration",
            "conf",
            required=True,
            type=click.Path(exists=True),
            help="Input configuration file to pass.",
        ),
        click.option(
            "-o",
            "--output",
            "output",
            required=False,
            type=click.Path(),
            default=None,
            help="Output file.",
        ),
        click.option(
            "--nThreads",
            "numberOfThreads",
            required=False,
            type=int,
            default=None,
            help="Number of threads for parallel processing.",
        ),
        click.option(
            "-d",
            "--debug",
            is_flag=True,
            help="Enable debug mode.",
        ),
        click.option(
            "-l",
            "--logger",
            is_flag=True,
            help="Save log file.",
        ),
        click.option(
            "-P",
            "--plot",
            is_flag=True,
            help="Save plots.",
        ),
    ]
    for option in reversed(options):  # Add options in correct order
        func = option(func)
    return func


@click.group(cls=RichGroup)
@click.version_option(
    version=__version__,
    prog_name="ExoSim",
    message="%(prog)s v%(version)s",
)
def cli():
    """ExoSim CLI - A simulation toolkit for exoplanet characterisation."""


@cli.command()
@common_options
@click.option(
    "--plot-scale",
    default="linear",
    type=click.Choice(["linear", "dB"], case_sensitive=False),
    show_default=True,
    help="Plot scale. Can be 'linear' or 'dB'.",
)
def focalplane(conf, output, numberOfThreads, debug, logger, plot, plot_scale):
    """Create and plot the focal plane."""
    _set_log(debug, logger, output)
    _set_threads(numberOfThreads)

    recipes.CreateFocalPlane(options_file=conf, output_file=output)

    if plot:
        _plot_focal_plane(output, plot_scale)


@cli.command()
@common_options
def radiometric(conf, output, numberOfThreads, debug, logger, plot):
    """Create a radiometric model."""
    _set_log(debug, logger, output)
    _set_threads(numberOfThreads)

    recipes.RadiometricModel(options_file=conf, input_file=output)

    if plot:
        _plot_radiometric(output)


@cli.command()
@common_options
@click.option(
    "-i",
    "--input",
    "input",
    required=True,
    type=click.Path(exists=True),
    help="Input file.",
)
@click.option(
    "--chunk-size",
    default=2,
    type=float,
    show_default=True,
    help="H5 file chunk size.",
)
def subexposures(
    conf, input, output, numberOfThreads, debug, logger, plot, chunk_size
):
    """Create and plot sub-exposures."""
    _set_log(debug, logger, output)
    _set_threads(numberOfThreads)
    RunConfig.chunk_size = chunk_size

    recipes.CreateSubExposures(
        options_file=conf, input_file=input, output_file=output
    )

    if plot:
        _plot_subexposures(output)


@cli.command()
@common_options
@click.option(
    "-i",
    "--input",
    "input",
    required=True,
    type=click.Path(exists=True),
    help="Input file.",
)
@click.option(
    "--chunk-size",
    default=2,
    type=float,
    show_default=True,
    help="H5 file chunk size.",
)
def ndrs(
    conf, input, output, numberOfThreads, debug, logger, plot, chunk_size
):
    """Create and plot NDRs."""
    _set_log(debug, logger, output)
    _set_threads(numberOfThreads)
    RunConfig.chunk_size = chunk_size

    recipes.CreateNDRs(options_file=conf, input_file=input, output_file=output)

    if plot:
        _plot_ndrs(output)


# Support functions for shared behaviour
def _set_log(debug, log, output):
    """Configure logging based on options."""
    if debug:
        setLogLevel(logging.DEBUG)
    if log:
        log_file = "exosim.log"
        if output:
            log_dir = os.path.dirname(output)
            log_file = os.path.join(log_dir, "exosim.log")
        try:
            addLogFile(fname=log_file)
        except PermissionError:
            addLogFile(fname="exosim.log")


def _set_threads(numberOfThreads):
    """Set the number of threads for parallel processing."""
    if numberOfThreads:
        RunConfig.n_job = numberOfThreads


# Plot functions for specific outputs
def _plot_focal_plane(output, scale):
    """Generate focal plane plots."""
    import exosim.plots as plots

    dir_name = os.path.join(os.path.dirname(output), "plots")
    os.makedirs(dir_name, exist_ok=True)

    focal_plane_plotter = plots.FocalPlanePlotter(input=output)
    focal_plane_plotter.plot_focal_plane(time_step=0, scale=scale)
    focal_plane_plotter.save_fig(os.path.join(dir_name, "focal_plane.png"))
    focal_plane_plotter.plot_efficiency()
    focal_plane_plotter.save_fig(os.path.join(dir_name, "efficiency.png"))


def _plot_radiometric(output):
    """Generate radiometric model plots."""
    import exosim.plots as plots

    dir_name = os.path.join(os.path.dirname(output), "plots")
    os.makedirs(dir_name, exist_ok=True)

    radiometric_plotter = plots.RadiometricPlotter(input=output)
    radiometric_plotter.plot_table()
    radiometric_plotter.save_fig(os.path.join(dir_name, "radiometric.png"))
    radiometric_plotter.plot_apertures()
    radiometric_plotter.save_fig(os.path.join(dir_name, "apertures.png"))


def _plot_subexposures(output):
    """Generate sub-exposures plots."""
    import exosim.plots as plots

    dir_name = os.path.join(os.path.dirname(output), "plots")
    os.makedirs(dir_name, exist_ok=True)

    sub_exposures_plotter = plots.SubExposuresPlotter(input=output)
    sub_exposures_plotter.plot(dir_name)


def _plot_ndrs(output):
    """Generate NDRs plots."""
    import exosim.plots as plots

    dir_name = os.path.join(os.path.dirname(output), "plots")
    os.makedirs(dir_name, exist_ok=True)

    ndrs_plotter = plots.NDRsPlotter(input=output)
    ndrs_plotter.plot(dir_name)


if __name__ == "__main__":
    cli()



from .applyAstronomicalSignal import ApplyAstronomicalSignal
from .estimateAstronomicalSignal import EstimateAstronomicalSignal
from .estimatePlanetarySignal import EstimatePlanetarySignal
from .findAstronomicalSignals import FindAstronomicalSignals



import os
from copy import deepcopy
from typing import List, Tuple

import h5py
import matplotlib
import matplotlib.colors as mcolors
import numpy as np


def _create_ordered_cmap(
    map_name: str,
    roll: int = None,
    delete: int = None,
    change: List[Tuple[int, int]] = None,
) -> mcolors.ListedColormap:
    """
    Create an ordered colormap based on a given colormap name.

    Parameters:
    - map_name (str): The name of the colormap to use.
    - roll (int, optional): The number of positions to roll the colormap colors. Defaults to None.
    - delete (int, optional): The index of the color to delete from the colormap. Defaults to None.
    - change (list[tuple[int, int]], optional): A list of pairs of indices to swap colors in the colormap. Defaults to None.

    Returns:
    - cmap (matplotlib.colors.ListedColormap): The created ordered colormap.
    """
    pastel1_cmap = matplotlib.colormaps.get_cmap(map_name)
    pastel1_colors = pastel1_cmap(np.linspace(0, 1, pastel1_cmap.N))

    pastel1_colors_hsv = np.array(
        [mcolors.rgb_to_hsv(color[:3]) for color in pastel1_colors]
    )
    sorted_indices = np.argsort(pastel1_colors_hsv[:, 0])
    sorted_colors = pastel1_colors[sorted_indices]

    reversed_colors = sorted_colors[::-1]
    if roll is not None:
        reversed_colors = np.roll(reversed_colors, roll, axis=0)
    if delete is not None:
        if isinstance(delete, List):
            for del_ in delete:
                reversed_colors = np.delete(reversed_colors, del_, axis=0)
        else:
            reversed_colors = np.delete(reversed_colors, delete, axis=0)

    if change is not None:
        for couple in change:
            val1 = deepcopy(reversed_colors[couple[0]])
            reversed_colors[couple[0]] = deepcopy(reversed_colors[couple[1]])
            reversed_colors[couple[1]] = val1

    cmap = mcolors.ListedColormap(reversed_colors)
    return cmap


def prepare_channels_list(
    input_file,
) -> Tuple[np.ndarray, matplotlib.colors.Normalize]:
    """
    Prepare the list of channels and the normalization object for plotting.

    Returns
    -------
    Tuple[np.ndarray, matplotlib.colors.Normalize]
        The sorted array of channel names and the normalization object.
    """
    with h5py.File(input_file, "r") as f:
        channels = np.array(list(f["channels"].keys()))
        channels_wl = []
        for channel_name in channels:
            file_path = os.path.join("channels", channel_name)
            file_path = os.path.join(file_path, "focal_plane/metadata")
            channels_wl.append(f[os.path.join(file_path, "wl_min/value")][()])
        id_ = np.argsort(np.array(channels_wl))
        channels = channels[id_]
        norm = matplotlib.colors.Normalize(vmin=0.0, vmax=len(channels))
    return channels, norm



import numpy as np
from astropy.io import ascii
from numba import jit, prange

from exosim.output import Output
from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks


class ApplyDeadPixelsMap(Task):
    """
    It masks the dead pixel in the array given their coordinates.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        outputs: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", " ")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("applying dead pixel map")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        output = self.get_task_param("output")

        self.model(subexposures, parameters, output)

    def model(self, subexposures, parameters, output):
        dead_pixels_map = np.ones(
            (subexposures.shape[1], subexposures.shape[2])
        )

        dead_coords = ascii.read(parameters["detector"]["dp_map"])

        if output:
            if issubclass(output.__class__, Output):
                out_grp = output.create_group("dead_pixel_map")
                out_grp.write_table("dead_pixel_map", dead_coords)

        for x, y in dead_coords["spectral_coords", "spatial_coords"]:
            dead_pixels_map[y, x] = 0.0

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="applying dead pixel map"
        ):
            subexposures.dataset[chunk] = self.add_dead_pixels(
                subexposures.dataset[chunk], dead_pixels_map
            )
            subexposures.output.flush()

    @staticmethod
    @jit(nopython=True, parallel=True)
    def add_dead_pixels(ndrs, dead_pixels_map):
        for t in prange(ndrs.shape[0]):
            ndrs[t] = ndrs[t] * dead_pixels_map
        return ndrs



import astropy.units as u
import numpy as np

from .checks import check_units


def wl_grid(wl_min, wl_max, R, return_bin_width=False):
    r"""
    It returns the wavelength log-grid in microns

    The wavelength at the center of the spectral bins is defined as

    .. math::

        \lambda_c = \frac{1}{2} (\lambda_j + \lambda_{j+1} )

    where :math:`\lambda_j` is the wavelength at the bin edge defined by the recursive relation,
    and :math:`R` is the `logbin_resolution` defined by the user.

    .. math::

        \lambda_{j+1} = \lambda_{j} \left( 1 + \frac{1}{R} \right)

    And, given the maximum and minimum wavelengths, provided by the user, the number of bins is

    .. math::

        n_{bins} = \frac{\log \left( \frac{\lambda_{max}}{\lambda_{min}} \right) } {\log \left( 1 + \frac{1}{R}\right)} + 1


    Parameters
    ----------
    wl_min: :class:`~astropy.units.Quantity` or float.
        minimum wavelength sampled. If no units are attached is considered as expressed in `um`
    wl_max: :class:`~astropy.units.Quantity` or float.
        maximum wavelength sampled. If no units are attached is considered as expressed in `um`
    R: int
        spectral resolving power
    return_bin_width: bool
        if True returns also the bin width. Default is False.

    Returns
    -------
    :class:`~astropy.units.Quantity`
        wavelength grid
    :class:`~astropy.units.Quantity`
        bin wavelength width
    """
    wl_min, wl_max = check_units(wl_min, u.um, force=True), check_units(
        wl_max, u.um, force=True
    )

    number_of_spectral_bins = (
        np.ceil(np.log(wl_max / wl_min) / np.log(1.0 + 1.0 / R)) + 1
    )
    wl_bin = wl_min * (1.0 + 1.0 / R) ** np.arange(number_of_spectral_bins)
    wl_bin_c = 0.5 * (wl_bin[0:-1] + wl_bin[1:])
    wl_bin_width = wl_bin[1:] - wl_bin[0:-1]

    if return_bin_width:
        return wl_bin_c, wl_bin_width
    else:
        return wl_bin_c


def time_grid(time_min, time_max, low_frequencies_resolution=None):
    """
    It returns the time grid in hours

    Parameters
    ----------
    time_min: :class:`~astropy.units.Quantity` or float.
        minimum time sampled. If no units are attached is considered as expressed in hours.
    time_max: :class:`~astropy.units.Quantity` or float.
        maximum time sampled. If no units are attached is considered as expressed in hours.
    low_frequencies_resolution: :class:`~astropy.units.Quantity`, float or None
        time sampling interval. If no units are attached is considered as expressed in hours.
        If None, a single time point at time_min is returned.

    Returns
    -------
    :class:`~astropy.units.Quantity`
        time grid
    """
    time_min = check_units(time_min, u.hr, force=True)
    time_max = check_units(time_max, u.hr, force=True)

    if low_frequencies_resolution is None:
        return np.array([time_min.value]) * u.hr

    low_frequencies_resolution = check_units(
        low_frequencies_resolution, u.hr, force=True
    )

    return (
        np.arange(
            time_min.value, time_max.value, low_frequencies_resolution.value
        )
        * u.hr
    )



import h5py

import exosim.output as output
from exosim.output.hdf5.utils import recursively_read_dict_contents
from exosim.tasks.task import Task


class LoadPixelsNonLinearityMap(Task):
    """
    Loads the pixels non-linearuty map

    Returns
    --------
    dict
        channel non linearity map

    Raises
    ------
    KeyError:
        if the output do not have the `map` key

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("loading pixels non-linearity map")
        parameters = self.get_task_param("parameters")

        pnl_dict_map = self.model(parameters)

        # checking output
        if "map" not in pnl_dict_map:
            self.error("map missing in pixel non-linearity map")
            raise KeyError("map missing in pixel non-linearity map")

        self.debug("pnl_variation_map: {}".format(pnl_dict_map["map"]))

        output_file = self.get_task_param("output")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                output_file.store_dictionary(
                    pnl_dict_map, "pixel_non_linearity"
                )

        self.set_output(pnl_dict_map)

    def model(self, parameters: dict) -> dict:
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        dict
            channel pixel non-linearity map

        """
        file_name = parameters["detector"]["pnl_filename"]
        ch_name = parameters["value"]

        with h5py.File(file_name, "r") as f:
            pnl_dict = recursively_read_dict_contents(f[ch_name])

        return pnl_dict



from .adcGainEstimator import ADCGainEstimator
from .darkCurrentMap import DarkCurrentMap
from .deadPixelsMap import DeadPixelsMap
from .pixelsNonLinearity import PixelsNonLinearity
from .pixelsNonLinearityFromCorrection import PixelsNonLinearityFromCorrection
from .quantumEfficiencyMap import QuantumEfficiencyMap
from .readoutSchemeCalculator import ReadoutSchemeCalculator



from .applyIntraPixelResponseFunction import ApplyIntraPixelResponseFunction
from .computeSaturation import ComputeSaturation
from .computeSolidAngle import ComputeSolidAngle
from .computeSourcesPointingOffset import ComputeSourcesPointingOffset
from .createFocalPlane import CreateFocalPlane
from .createFocalPlaneArray import CreateFocalPlaneArray
from .createIntrapixelResponseFunction import CreateIntrapixelResponseFunction
from .createOversampledIntrapixelResponseFunction import (
    CreateOversampledIntrapixelResponseFunction,
)
from .foregroundsToFocalPlane import ForegroundsToFocalPlane
from .loadPsf import LoadPsf
from .loadPsfPaos import LoadPsfPaos
from .loadPsfPaosTimeInterp import LoadPsfPaosTimeInterp
from .loadResponsivity import LoadResponsivity
from .loadWavelengthSolution import LoadWavelengthSolution
from .populateFocalPlane import PopulateFocalPlane
from .propagateForegrounds import PropagateForegrounds
from .propagateSources import PropagateSources



from copy import copy

import astropy.units as u
import numpy as np

from exosim.models.signal import CountsPerSecond, Signal
from exosim.tasks.task import Task
from exosim.utils.checks import check_units


class ComputeSaturation(Task):
    """
    It computes the saturation time given the focal plane and the detector parameters.

    Returns
    --------
    :class:`~astropy.units.Quantity`
        saturation time
    :class:`~astropy.units.Quantity`
        frame time
    :class:`~astropy.units.Quantity`
        maximum signal
    :class:`~astropy.units.Quantity`
        minimum signal
    """

    def __init__(self):
        """
        Parameters
        __________
        well_depth: :class:`~astropy.units.Quantity`
            detector well depth
        f_well_depth: float
            fraction of detector well depth to use
        focal_plane: :class:`~astropy.units.Quantity`
            focal plane array (with time evolution)
        frg_focal_plane: :class:`~astropy.units.Quantity`
            foreground focal plane array (with time evolution)

        """
        self.add_task_param("well_depth", "detector well depth")
        self.add_task_param(
            "f_well_depth", "fraction of detector well depth to use", None
        )
        self.add_task_param("focal_plane", "focal plane")
        self.add_task_param("frg_focal_plane", "foreground focal plane", None)

    def execute(self):
        well_depth = self.get_task_param("well_depth")
        f_well_depth = self.get_task_param("f_well_depth")
        focal_plane = self.get_task_param("focal_plane")
        frg_focal_plane = self.get_task_param("frg_focal_plane")

        well_depth = check_units(
            well_depth, desired_units=u.ct, calling_class=self
        )

        if isinstance(focal_plane, (Signal, CountsPerSecond)):
            focal_plane = focal_plane.data * focal_plane.data_units
        else:
            focal_plane = check_units(
                focal_plane, desired_units=u.ct / u.s, calling_class=self
            )

        if frg_focal_plane is not None:
            if isinstance(frg_focal_plane, (Signal, CountsPerSecond)):
                frg_focal_plane = (
                    frg_focal_plane.data * frg_focal_plane.data_units
                )
            else:
                frg_focal_plane = check_units(
                    frg_focal_plane,
                    desired_units=u.ct / u.s,
                    calling_class=self,
                )
            focal_plane += frg_focal_plane

        max_signal = np.max(focal_plane)
        min_signal = np.min(focal_plane)

        saturation_time = well_depth / max_signal
        self.debug("saturation time : {}".format(saturation_time))

        integration_time = copy(saturation_time)
        if f_well_depth is not None:
            integration_time *= f_well_depth
        self.debug("integration_time time : {}".format(integration_time))

        self.set_output(
            [saturation_time, integration_time, max_signal, min_signal]
        )



import importlib.util
import json

import astropy.units as u
import numpy as np
from astropy.io import ascii
from joblib import Parallel, delayed
from scipy.interpolate import interp1d
from tqdm.auto import tqdm

from exosim.tasks.astrosignal.estimateAstronomicalSignal import (
    EstimateAstronomicalSignal,
)
from exosim.utils import RunConfig
from exosim.utils.binning import rebin
from exosim.utils.checks import check_units, find_key
from exosim.utils.types import ArrayType

batman_spec = importlib.util.find_spec("batman")
if batman_spec is not None:
    import batman


class EstimatePlanetarySignal(EstimateAstronomicalSignal):
    """
    This task estimates the planetary signal.
    The signal parameters are extracted from the sky configuration file.
    This Task is a wrapper for the batman package: it parses the planet parameters and computes the transit model.
    The transit model is computed at the input wavelength grid.
    By default, the transit type is set to 'primary' (i.e. the planet transits in front of the star).
    The planetary sarius can be parsed as a string (path to a file) or as a float (in units of stellar radii).
    If a file is indicated, the planetary radius is binned or interpolated to the input wavelength grid.

    Based on Batman package by L. Kreidberg (https://ui.adsabs.harvard.edu/abs/2015PASP..127.1161K/abstract).
    """

    def model(
        self,
        timeline: ArrayType,
        wl_grid: ArrayType,
        ch_parameters: dict = {},
        source_parameters: dict = {},
    ) -> ArrayType:
        """
        Parameters
        ----------
        timeline : :class:`astropy.units.Quantity`
            timeline to compute the signal
        wl_grid : :class:`astropy.units.Quantity`
            wavelength grid
        ch_parameters : dict, optional
            channel parameters, by default {}
        source_parameters : dict
            source parameters, by default {}

        Returns
        -------
        ArrayType
            returns the planetary signal in a 2D array (wavelength, time)
        """

        self.info("Estimating planetary signal")
        # set timeline units to sec
        input_timeline = check_units(timeline, u.s, self, True)

        # load planetary parameters from sky parameters dictionary
        planet = source_parameters["planet"]
        params = batman.TransitParams()
        params.t0 = check_units(
            planet["t0"], input_timeline.unit, self, True
        ).value
        params.per = check_units(
            planet["period"], input_timeline.unit, self, True
        ).value
        params.a = planet["sma"]  # (in units of stellar radii)
        params.inc = check_units(planet["inc"], u.deg, self, True).value
        params.ecc = planet["ecc"]
        params.w = check_units(planet["w"], u.deg, self, True).value
        params.limb_dark = planet["limb_darkening"]
        transittype = (
            planet["transittype"] if "transittype" in planet else "primary"
        )

        # load planet rp
        rp_ = planet["rp"]
        if isinstance(rp_, str):
            self.debug("loading rp from file")
            rp = self.load_rp(rp_, wl_grid)
        else:
            try:
                self.debug("rp is a float")
                rp_ = float(rp_)
                rp = rp_ * np.ones(wl_grid.size)
            except ValueError:
                self.error("rp must be a string or a float")

        # load limb darkening coefficients
        ldc_ = planet["limb_darkening_coefficients"]
        try:
            self.debug("loading ldc from file")
            ldc = self.load_ldc(ldc_, wl_grid)
        except FileNotFoundError:
            try:
                self.debug("ldc is a list")
                ldc = json.loads(ldc_)
                ldc = np.array(ldc)
            except ValueError:
                self.error("ldc must be a string or a list")
                raise ValueError(
                    "ldc must be a string (path to a file) or a list of coefficients"
                )

        ldc = np.array(ldc)

        self.debug("transit center: {}".format(params.t0))
        # reduce the timeline to the transit duration
        transit_durations = self.get_t14(
            params.inc * u.deg,
            params.a,
            params.per * input_timeline.unit,
            rp,  # use max rp to maximise the transit duration
        )

        margins = 1.05  # apply a 5% margin to the transit duration to avoid edge effects
        self.debug("marging to transit duration: {}".format(margins))
        transit_durations *= margins
        max_transit_duration = np.nanmax(transit_durations)
        self.debug("max transit durations: {}".format(transit_durations))
        start_t = (
            np.abs(
                input_timeline
                - (params.t0 * input_timeline.unit - max_transit_duration / 2)
            )
        ).argmin()
        self.debug(
            "starting model at: {}".format(
                params.t0 * input_timeline.unit - max_transit_duration / 2
            )
        )
        end_t = (
            np.abs(
                input_timeline
                - (params.t0 * input_timeline.unit + max_transit_duration / 2)
            )
        ).argmin()
        self.debug(
            "ending model at: {}".format(
                params.t0 * input_timeline.unit + max_transit_duration / 2
            )
        )

        self.debug(
            "Reducing timeline to the transit duration: {} - {}".format(
                start_t, end_t
            )
        )
        timeline = input_timeline[start_t:end_t]
        self.debug("edited timeline size: {}".format(timeline.size))

        # iterate over the wavelength grid
        self.debug("computing the transit model")
        out_model_list = Parallel(n_jobs=RunConfig.n_job, require="sharedmem")(
            delayed(self._compute_transit_model)(
                i, rp, ldc, timeline, params, transittype
            )
            for i in tqdm(
                range(wl_grid.size),
                total=wl_grid.size,
                desc="building transit model over wavelengths",
            )
        )
        out_model = np.array(out_model_list)

        return timeline, out_model

    def load_rp(self, path: str, wl: ArrayType) -> np.ndarray:
        """
        Load the planet radius from a file and rebin it to the input wavelength grid.

        Parameters
        ----------
        path : str
            Path to the file containing the planet radius data.
            The file must include columns named "Wavelength" and "rp/rs" representing the wavelengths
            and the corresponding planet radius (in units of stellar radii), respectively.
        wl : astropy.units.Quantity
            Wavelength grid onto which the planet radius data will be interpolated.

        Returns
        -------
        np.ndarray
            Array of planet radius values (rp/rs) interpolated onto the specified wavelength grid.

        Notes
        -----
        The function reads the data using `ascii.read` and then applies linear interpolation via
        `interp1d` to bin the planet radius values to the provided wavelength grid. Extrapolation is
        enabled for wavelengths that fall outside the range of the data.
        """
        rp_data = ascii.read(path)
        self.debug("binning rp to the input wavelength grid")
        rp_data["Wavelength"] = check_units(
            rp_data["Wavelength"], wl.unit, self, True
        ).value

        rp = rebin(
            wl,
            rp_data["Wavelength"],
            rp_data["rp/rs"],
            fill_value="extrapolate",
        )

        return rp

    def load_ldc(self, path: str, wl: ArrayType) -> np.ndarray:
        """
        Load the limb darkening coefficients from a file and rebin them to the input wavelength grid.

        Parameters
        ----------
        path : str
            Path to the file containing the limb darkening coefficients data.
            The file must include a column representing the wavelength (e.g. "wavelength" or "wl") and one or more
            columns with the corresponding limb darkening coefficients.
        wl : astropy.units.Quantity
            The wavelength grid to which the limb darkening coefficients will be rebinned.

        Returns
        -------
        np.ndarray
            An array of rebinned limb darkening coefficients. If the file contains multiple coefficient columns,
            the resulting array will be 2-dimensional with shape (n_coeff, len(wl)), where n_coeff is the number
            of coefficient columns.

        Notes
        -----
        The function utilises `ascii.read` to load the data from the file.
        The wavelength column is identified using the `find_key` function by searching for keys such as "wavelength" or "wl".
        Each remaining column is rebinned to the provided wavelength grid using the `rebin` function.
        """
        ldc_data = ascii.read(path)

        wl_key = find_key(ldc_data.keys(), ["wavelength", "wl"], self)
        ldc_wl = ldc_data[wl_key]
        ldc_wl = check_units(ldc_wl, wl.unit, self, True)

        ldc_out = []
        self.debug("binning ldc to the input wavelength grid")
        for k in ldc_data.keys():
            if "ldc" in k:
                ldc = rebin(wl, ldc_wl, ldc_data[k], fill_value="extrapolate")
                ldc_out.append(ldc)
        return ldc_out

    def get_t14(
        self, inc: u.Quantity, sma: float, period: u.Quantity, rp: np.ndarray
    ) -> u.Quantity:
        """t14
        Calculates the transit time based on the work of Seager, S., & Mallen-Ornelas, G. 2003, ApJ, 585, 1038

        Parameters
        __________
        inc: :class:`astropy.units.Quantity`
            Planet oprbital inclination
        sma: float
            Semimajor axis in stellar units
        period: :class:`astropy.units.Quantity`
            Orbital period
        rp: np.ndarray
            Planet radius in stellar units

        Returns
        __________
        transit duration : :class:`astropy.units.Quantity`
            Returns the transit duration
        """
        period = period.to(u.s)
        impact_parameter = np.cos(inc.to(u.rad)) * sma
        t14 = (
            period / np.pi / sma * np.sqrt((1 + rp) ** 2 - impact_parameter**2)
        )
        self.debug("transit duration: {}".format(t14))
        return t14

    def _compute_transit_model(
        self, i, rp, ldc, timeline, params, transittype
    ):
        """
        Compute the transit model for a single wavelength index.

        Parameters
        ----------
        i : int
            Wavelength index.
        rp : np.ndarray
            Array of planet radius values for each wavelength.
        ldc : np.ndarray
            Array of limb darkening coefficients.
        timeline : astropy.units.Quantity
            Timeline reduced to the transit duration.
        params : batman.TransitParams
            Transit parameters.
        transittype : str
            Type of transit (e.g. 'primary').

        Returns
        -------
        np.ndarray
            The computed light curve for the given wavelength.
        """
        params.rp = rp[i]  # Set planet radius for the current wavelength
        if ldc.ndim == 2:
            params.u = list(ldc[:, i])
        else:
            params.u = ldc
        # Initialise batman model and compute light curve
        m = batman.TransitModel(params, timeline, transittype=transittype)
        return m.light_curve(params)



from copy import deepcopy

import h5py
import numpy as np

from exosim.models.signal import Counts
from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks
from exosim.utils.types import HDF5OutputType


class ApplyPixelsNonLinearity(Task):
    r"""
    Given the pixel non-linearity parameters,
    this Task correct the ideal measured signal rate to the pixel linearity.

    .. math::
        Q_{det} = Q \cdot (a + b \cdot Q + c \cdot Q^2 + d \cdot Q^3 + e \cdot Q^4)

    The input is a dictionary with a `map` keyword containg an array with the coefficients for each pixel.
    The map shape is (n_pixels_x, n_pixels_y, coefficient order).

    The user can list any number of coefficients, that will be parsed in the following model

    .. math::
        Q_{det} = Q \cdot (a_0 + \sum_i a_i \cdot Q^i)

    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        """

        self.add_task_param("subexposures", " ")
        self.add_task_param("parameters", "channel non linearity dictionary")

    def execute(self):
        self.info("applying pixel non-linearity map")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")

        self.model(subexposures, parameters)

    def model(self, subexposures: Counts, parameters: dict) -> None:
        pnl_map = parameters["map"][:]

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="applying pixel non-linearity map"
        ):
            data = deepcopy(subexposures.dataset[chunk])
            npl = 0
            for i in range(0, len(pnl_map)):
                npl += data ** (i + 1) * np.repeat(
                    pnl_map[np.newaxis, i], data.shape[0], axis=0
                )
            subexposures.dataset[chunk] = npl
            subexposures.output.flush()



import astropy.units as u
import numpy as np

from exosim.tasks.instrument.createIntrapixelResponseFunction import (
    CreateIntrapixelResponseFunction,
)


class CreateOversampledIntrapixelResponseFunction(
    CreateIntrapixelResponseFunction
):
    def model(self, parameters):
        """
        This class produces an oversampled version of the intrapixel response function.
        This kernel is zero-padded to the size of the PSF.
        This version is compatible with :func:`exosim.utils.convolution.fast_convolution` convolution function.


        Estimate the detector pixel response function with the prescription of
        Barron et al., PASP, 119, 466-475 (2007).

        Parameters
        ----------
        oversampling: int
                number of samples in each resolving element. The
                final shape of the response function would be shape*osf
        delta_pix: :class:`astropy.units.Quantity`
                Phisical size of the detector pixel in microns
        diffusion_length: :class:`astropy.units.Quantity`
                diffusion length in microns
        intra_pix_distance: :class:`astropy.units.Quantity`
                distance between two adjacent detector pixels
                in microns

        Returns
        -------
        kernel	: 2D array
                the kernel image
        kernel_delta  : :class:`astropy.units.Quantity`
                        the kernel sampling interval in microns
        """
        if "oversampling" in parameters["detector"].keys():
            osf = (
                8 * parameters["detector"]["oversampling"]
            )  # activate for old convolution
        else:
            osf = 7  # activate for old convolution

        # TODO find a better zeroing method. It should be enought if the zeros are twice the size of the kernel.
        psf_shape = parameters["psf_shape"]

        delta = parameters["detector"]["delta_pix"]
        if "diffusion_length" in parameters["detector"].keys():
            lx = parameters["detector"]["diffusion_length"]
        else:
            lx = 0.0 * u.um
        if "intra_pix_distance" in parameters["detector"].keys():
            ipd = parameters["detector"]["intra_pix_distance"]
        else:
            ipd = 0 * u.um

        if type(osf) != int:
            osf = int(osf)

        lx += 1e-20 * u.um  # to avoid problems if user pass lx=0
        lx = lx.to(delta.unit)

        kernel = np.zeros((psf_shape[0] * osf, psf_shape[1] * osf))
        self.debug("kernel size: {}".format(kernel.shape))

        # prepare the kernel stamp grid
        kernel_delta = delta / osf
        yc, xc = np.array(kernel.shape) // 2
        yy = (np.arange(-1 * osf, 1 * osf)) * kernel_delta
        xx = (np.arange(-1 * osf, 1 * osf)) * kernel_delta

        # inverse mask to select every other pixel but the central one
        i_mask_xx = np.where(np.abs(xx) > 0.5 * (delta - ipd))
        i_mask_yy = np.where(np.abs(yy) > 0.5 * (delta - ipd))

        xx, yy = np.meshgrid(xx, yy)

        # compute kernel stamp
        kernel_stamp = np.arctan(
            np.tanh((0.5 * (0.5 * delta - xx) / lx).value)
        ) - np.arctan(np.tanh((0.5 * (-0.5 * delta - xx) / lx).value))

        kernel_stamp *= np.arctan(
            np.tanh((0.5 * (0.5 * delta - yy) / lx).value)
        ) - np.arctan(np.tanh((0.5 * (-0.5 * delta - yy) / lx).value))

        # set the unused area of kernel stamp to zero
        kernel_stamp[i_mask_yy, ...] = 0.0
        kernel_stamp[..., i_mask_xx] = 0.0

        # Normalise the kernel such that the pixel has QE=1
        kernel_stamp *= osf * osf / kernel_stamp.sum()
        # put back the kernel stamp
        x_off = kernel.shape[1] // 2 - kernel_stamp.shape[1] // 2
        y_off = kernel.shape[0] // 2 - kernel_stamp.shape[0] // 2

        kernel[
            y_off : y_off + kernel_stamp.shape[0],
            x_off : x_off + kernel_stamp.shape[1],
        ] = kernel_stamp

        # roll to use for the fast fft convolve
        kernel = np.roll(kernel, -xc, axis=1)
        kernel = np.roll(kernel, -yc, axis=0)

        return kernel, kernel_delta

        # Author note: previous version was computing the pixel irf shape on the full kernel and then setting to 0 all other pixels.
        # That method resulted in a very slow computation of arctan and huge memory requiring for high osf.
        # This "kernel stamp" solution return the same kernel (with relative differences under 1e-8 - tested on 2022/05/19)
        # but very lower computing time.
        # TEST: using osf 12 psf =(64,64): previous method took 4.2s, new method takes 6ms with)
        # This method removes the memory problem: not much of RAM used for this computation
        # tested on jupyter notebook on 2022/05/19. L.V.M.



from exosim.output import output
from exosim.tasks.task import Task
from exosim.utils.types import ArrayType


class EstimateAstronomicalSignal(Task):
    """
    It is a base class for all astronomical signal estimation tasks.
    If an output file is provided, the signal model is stored in the output file.

    Returns
    -------
    ArrayType
        returns the planetary signal in a 2D array (wavelength, time)
    """

    def __init__(self) -> None:
        """
        Parameters
        ----------
        timeline : :class:`astropy.units.Quantity`
            timeline to compute the signal
        wl_grid : :class:`astropy.units.Quantity`
            wavelength grid
        ch_parameters : dict
            channel parameters, by default {}
        source_parameters : dict
            sourec parameters, by default {}
        output : :class:`exosim.output.output.Output` (optional)
            output file, by default None
        """

        self.add_task_param("timeline", "timeline")
        self.add_task_param("source_parameters", "")
        self.add_task_param("ch_parameters", "")
        self.add_task_param("wl_grid", "")
        self.add_task_param("output", "", None)

    def execute(self):
        timeline = self.get_task_param("timeline")
        source_parameters = self.get_task_param("source_parameters")
        ch_parameters = self.get_task_param("ch_parameters")
        wl_grid = self.get_task_param("wl_grid")

        new_timeline, model = self.model(
            timeline, wl_grid, ch_parameters, source_parameters
        )

        # store the signal model in the output file
        output_file = self.get_task_param("output")
        if output_file and issubclass(output_file.__class__, output.Output):
            output_file.write_array(
                "signal_model",
                model,
                metadata={
                    "wavelength_axis": 0,
                    "time_axis": 1,
                },
            )
            output_file.write_quantity("model_timeline", new_timeline)
            output_file.write_quantity("model_wavelength", wl_grid)
            output_file.store_dictionary(source_parameters, "parameters")

        self.set_output([new_timeline, model])

    def model(
        self,
        timeline: ArrayType,
        wl_grid: ArrayType,
        ch_parameters: dict = {},
        source_parameters: dict = {},
    ) -> ArrayType:
        """Astronomical signal model to implement.

        Parameters
        ----------
        timeline : :class:`astropy.units.Quantity`
            timeline to compute the signal
        wl_grid : :class:`astropy.units.Quantity`
            wavelength grid
        ch_parameters : dict, optional
            channel parameters, by default {}
        source_parameters : dict
            source parameters, by default {}

        Returns
        -------
        ArrayType
            returns the planetary signal in a 2D array (wavelength, time)

        Raises
        ------
        NotImplementedError
            if the model is not implemented
        """
        raise NotImplementedError



import exosim.output as output
import exosim.utils.klass_factory as klass_factory
from exosim.tasks.load import LoadOpticalElement
from exosim.tasks.task import Task


class ParseOpticalElement(Task):
    """
    Given the element parameters, it parses the optical element and returns a dictionary.
    It also applyes the time variation if provided.

    Returns
    -------
    dict
        dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
        represeting the radiance and efficiency of the optical element.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        self.add_task_param("parameters", "optical element parameters dict")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file", None)

    def execute(self):
        parameters = self.get_task_param("parameters")
        self.debug("parsing: {}".format(parameters["value"]))

        wl = self.get_task_param("wavelength")
        tt = self.get_task_param("time")

        element_task = (
            klass_factory.find_task(
                parameters["task_model"], LoadOpticalElement
            )
            if "task_model" in parameters.keys()
            else LoadOpticalElement
        )
        task_instance = element_task()
        radiance, efficiency = task_instance(
            parameters=parameters, wavelength=wl, time=tt
        )

        output_file = self.get_task_param("output")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                og_ = output_file.create_group(parameters["value"])
                radiance.write(og_, "radiance")
                efficiency.write(og_, "efficiency")

        out = {"radiance": radiance, "efficiency": efficiency}

        self.set_output(out)



import os
from typing import Tuple

import astropy.units as u
import h5py
import matplotlib.pyplot as plt
import numpy as np

import exosim.log as log
from exosim.utils.ascii_arts import observatory


class NDRsPlotter(log.Logger):
    """
    Sub-Exposures plotter.
    This class handles the methods to plot all the sub-exposures produced by `ExoSim`.


    Examples
    ----------
    The following example, given the `test_file.h5` preoduced by Exosim,
    plots the sub-exposures stores the figures in the indicated folder.

    >>> from exosim.plots import SubExposuresPlotter
    >>> subExposuresPlotter = SubExposuresPlotter(input='./test_ndr.h5')
    >>> subExposuresPlotter.plot('plots/')

    """

    def __init__(self, input: str) -> None:
        """

        Parameters
        ----------
        input: str
            input data
        """
        self.set_log_name()
        self.graphics(observatory)
        self.input = input
        self.announce("started")

    def plot(self, out_dir: str) -> None:
        """It iterates over the channels and plot the ndrs.

        Parameters
        ----------
        out_dir : str
            output directory
        """ """ """
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        with h5py.File(self.input, "r") as f:
            ch_list = list(f["channels"].keys())
            ch_list.sort()

            for ch in ch_list:
                self.info("plotting {}".format(ch))
                exposures, time_line, n_groups_per_ramp = self.load_ndrs(ch, f)

                for i in range(n_groups_per_ramp):
                    self.plot_NDRs(exposures, time_line, i, ch, out_dir)

    def plot_NDRs(
        self,
        ndrs: np.ndarray,
        time_line: u.Quantity,
        i: int,
        ch: str,
        out_dir: str,
    ) -> None:
        """It plots the ndrs for a given channel.

        Parameters
        ----------
        ndrs : np.ndarray
            ndrs array
        time_line : u.Quantity
            temporal array
        i : int
            index of ndr to plot
        ch : str
            channel name
        out_dir : str
            output directory name
        """
        t = time_line[i]

        size_x = ndrs[i].shape[1]
        size_y = ndrs[i].shape[0]

        size_y_fig = 10
        size_x_fig = (size_y_fig * size_y / size_x) * 1.2

        fig, ax0 = plt.subplots(
            constrained_layout=True, dpi=300, figsize=(size_y_fig, size_x_fig)
        )

        if t < 0.01 * u.hr:
            t = t.to(u.s)

        t_str = "{:.3f} {}".format(t.value, t.unit)
        if t < 0.01 * u.s:
            t_str = "{:.2e} {}".format(t.value, t.unit)

        ax0.set_title("{} - num {} - time {}".format(ch, i, t_str))

        im = ax0.imshow(ndrs[i], interpolation="none")
        plt.colorbar(im, ax=ax0)

        fname = os.path.join(out_dir, "{}_ndrs_{}.png".format(ch, i))
        fig.savefig(fname, format="png", dpi=300)

        self.info("plot saved in {}".format(i))

    def load_ndrs(
        self, ch: str, f: h5py.File
    ) -> Tuple[np.ndarray, u.Quantity, int]:
        """
        It loads the channel NDRs from the input file:

        Parameters
        ----------
        ch: str
            channel name
        file: :class:`h5py.File`
           input file

        Returns
        -------
        :class:`numpy.ndarray`
            NDR
        :class:`astropy.units.Quantity`
            time line
        int
            number of groups per ramp
        """
        file_path = os.path.join("channels/", ch)
        se = f[os.path.join(file_path, "NDRs")]
        ndrs = se["data"]

        time_line = se["time"][()] * u.Unit(
            se["time_units"][()].decode("utf-8")
        )

        n_groups_per_ramp = se["metadata"]["n_groups_per_ramp"][()]

        self.debug("NDRs loaded from {}".format(ch))

        return ndrs, time_line, n_groups_per_ramp



import numpy as np
from astropy import units as u
from astropy.modeling.physical_models import BlackBody

import exosim.models.signal as signal
import exosim.utils.checks as checks
from exosim.tasks.task import Task


class CreatePlanckStar(Task):
    r"""
    Create a star SED using the Planck function.

    The star emission is simulated by :class:`astropy.modeling.physical_models.BlackBody`.
    The resulting sed is then converted into :math:`W/m^2/sr/\mu m` and scaled by the solid angle :math:`\pi \left( \frac{R}{D} \right)^2`.


    Returns
    -------
    :class:`~exosim.models.signal.Sed`
        Star Sed

    Examples
    --------
    >>> from exosim.tasks.sed import CreatePlanckStar
    >>> import astropy.units as u
    >>> import numpy as np
    >>> createPlanckStar = CreatePlanckStar()
    >>> wl = np.linspace(0.5, 7.8, 10000) * u.um
    >>> T = 6086 * u.K
    >>> R = 1.18 * u.R_sun
    >>> D = 47 * u.au
    >>> sed = createPlanckStar(wavelength=wl, T=T, R=R, D=D)

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(sed.spectral, sed.data[0,0])
    >>> plt.ylabel(sed.data_units)
    >>> plt.xlabel(sed.spectral_units)
    >>> plt.show()

    .. plot:: mpl_examples/createPlanckStar.py
    """

    def __init__(self):
        """
        Parameters
        __________
        wavelength: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
            wavelength grid. If no units are attached is considered as expressed in `um`
        T: :class:`~astropy.units.Quantity` or float
            star temperature. If no units are attached is considered as expressed in `K`
        R: :class:`~astropy.units.Quantity` or float
            star radius. If no units are attached is considered as expressed in `m`
        D: :class:`~astropy.units.Quantity` or float
            star distance. If no units are attached is considered as expressed in `m`
        """
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("T", "star temperature")
        self.add_task_param("R", "star radius")
        self.add_task_param("D", "star distannce")

    def execute(self):
        wl = self.get_task_param("wavelength")
        T = self.get_task_param("T")
        R = self.get_task_param("R")
        D = self.get_task_param("D")

        wl = checks.check_units(wl, u.um, self)
        T = checks.check_units(T, u.K, self)
        R = checks.check_units(R, u.m, self)
        D = checks.check_units(D, u.m, self)

        omega_star = np.pi * (R / D) ** 2 * u.sr
        bb = BlackBody(T)
        bb_ = bb(wl).to(u.W / u.m**2 / u.sr / u.um, u.spectral_density(wl))
        sed = signal.Sed(spectral=wl, data=omega_star * bb_)

        self.debug("star plack source created: {}".format(sed.data))

        self.set_output(sed)



import numpy as np
from astropy.table import QTable
from scipy.interpolate import interp1d

import exosim.utils.aperture as aperture
from exosim.tasks.task import Task


class EstimateApertures(Task):
    """
    It returns the sizes of the apertures to perform photometry for :func:`photutils.aperture.aperture_photometry`.
    The details of the apertures depends on the configurations set by the user.

    Returns
    --------
    astropy.table.QTable:

    Raises
    --------
    TypeError:
        if the output is not :class:`~astropy.table.QTable`

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        --------------
        table: :class:`astropy.table.QTable`
            wavelength table with bin edge
        focal_plane:
            focal plane
            Default is 'wl_solution`.
        wl_grid: (optional)
            focal plane wavelength solution
        description: dict (optional)
            channel description


        """
        self.add_task_param("table", "wavelength table with bin edges")
        self.add_task_param("focal_plane", "focal plane")
        # self.add_task_param('spatial_mode',
        #                     'aperture in spatial direction',
        #                     'column')
        # self.add_task_param('spectral_mode',
        #                     'aperture in spectral direction',
        #                     'wl_solution')
        self.add_task_param("wl_grid", "focal plane wavelength solution", None)
        self.add_task_param("description", "channel description", None)
        self.add_task_param("auto_mode", "automatic aperture mode", None)

    def execute(self):
        self.debug("performing aperture photometry")
        table = self.get_task_param("table")
        wl_grid = self.get_task_param("wl_grid")
        focal_plane = self.get_task_param("focal_plane")
        description = self.get_task_param("description")

        new_tab = self.model(table, focal_plane, description, wl_grid)

        if not isinstance(new_tab, QTable):
            self.error("wrong output format")
            raise TypeError("wrong output format")

        self.set_output(new_tab)

    def model(self, table, focal_plane, description, wl_grid):
        """
        It returns the sizes of the apertures to perform photometry for :func:`photutils.aperture.aperture_photometry`.
        The default methods to estimate the aperture are the following and these are also the keywords required for the `description` dictionary.

        + spatial_mode: str
            aperture in spatial direction. If `column` the aperture spatial length is the full pixel column in the array.
            Default is `column`.
        + spectral_mode: str
            aperture in spectral direction. If `row` the aperture is spectral length is the full pixel row of the array.
            If `wl_solution` the spectral size is the same of the spectral bin in the input table.
            Default is 'wl_solution`.
        + auto_mode: str
            automatic aperture mode. The dictionary must contains two keywords.
            The `mode` keyword should be a str: if `rectangular` then :func:`~exosim.utils.psf.find_rectangular_aperture` is used;
            if `elliptical` then :func:`~exosim.utils.psf.find_elliptical_aperture` is used;
            if `bin` then :func:`~exosim.utils.psf.find_bin_aperture` is used;
            if `full` then the full array is integrated.
        + Ene: float
            The `EnE` keyword should be a float and represents the Encircled Energy to include in the aperture.


        Parameters
        --------------
        table: :class:`astropy.table.QTable`
            wavelength table with bin edge
        focal_plane:
            focal plane
        description: dict
            dictionary containing the aperture photometry description

        Returns
        --------
        astropy.table.QTable:
        """

        # spectral settings
        if "spectral_mode" in description.keys():
            if description["spectral_mode"] == "row":
                self.debug("spectral mode: row")
                wlc = [focal_plane.shape[1] / 2] * table["Wavelength"].size
                dwl = [focal_plane.shape[1]] * table["Wavelength"].size
                shape = "rectangular"

            elif description["spectral_mode"] == "wl_solution":
                self.debug("computing wavelength solution")

                if wl_grid is None:
                    self.error("focal plane wavelength solution missing")
                    raise OSError("focal plane wavelength solution missing")

                wl_sol = interp1d(
                    wl_grid,
                    np.arange(0, focal_plane.shape[1]),
                    fill_value="extrapolate",
                )
                wlc = wl_sol(table["Wavelength"])
                wld = wl_sol(table["left_bin_edge"])
                wlu = wl_sol(table["right_bin_edge"])
                dwl = abs(wlu - wld)
                shape = "rectangular"

            else:
                self.error("not supported spectral mode")
                raise OSError("not supported spectral mode")

        # spatial settings
        if "spatial_mode" in description.keys():
            if description["spatial_mode"] == "column":
                self.debug("spatial mode: column")
                spc = [focal_plane.shape[0] / 2] * table["Wavelength"].size
                dsp = [focal_plane.shape[0]] * table["Wavelength"].size
                shape = "rectangular"
            else:
                self.error("not supported spatial mode")
                raise OSError("not supported spatial mode")

        # automatic settings
        if "auto_mode" in description.keys():
            if description["auto_mode"] == "full":
                wlc = [focal_plane.shape[1] / 2] * table["Wavelength"].size
                dwl = [focal_plane.shape[1]] * table["Wavelength"].size
                spc = [focal_plane.shape[0] / 2] * table["Wavelength"].size
                dsp = [focal_plane.shape[0]] * table["Wavelength"].size
                shape = "rectangular"

            if description["auto_mode"] == "elliptical":
                ene = description["EnE"]
                sizes, surf, ene = aperture.find_elliptical_aperture(
                    focal_plane, ene
                )
                wlc = [focal_plane.shape[1] / 2] * table["Wavelength"].size
                dwl = [sizes[0]] * table["Wavelength"].size
                spc = [focal_plane.shape[0] / 2] * table["Wavelength"].size
                dsp = [sizes[1]] * table["Wavelength"].size
                shape = "elliptical"

            if description["auto_mode"] == "rectangular":
                ene = description["EnE"]
                sizes, surf, ene = aperture.find_rectangular_aperture(
                    focal_plane, ene
                )
                wlc = [focal_plane.shape[1] / 2] * table["Wavelength"].size
                dwl = [sizes[1]] * table["Wavelength"].size
                spc = [focal_plane.shape[0] / 2] * table["Wavelength"].size
                dsp = [sizes[0]] * table["Wavelength"].size
                shape = "rectangular"

            if description["auto_mode"] == "bin":
                ene = description["EnE"]

                if wl_grid is None:
                    self.error("focal plane wavelength solution missing")
                    raise OSError("focal plane wavelength solution missing")

                wl_sol = interp1d(
                    wl_grid,
                    np.arange(0, focal_plane.shape[1]),
                    fill_value="extrapolate",
                )
                wlc = wl_sol(table["Wavelength"])
                wld = wl_sol(table["left_bin_edge"])
                wlu = wl_sol(table["right_bin_edge"])
                dwl = abs(wlu - wld)
                spc = [focal_plane.shape[0] / 2] * table["Wavelength"].size
                dsp = [1] * table["Wavelength"].size
                shape = "rectangular"

                self.debug("estimating apertures for spectral bins")
                for i in range(wlc.size):
                    size, surf, ene = aperture.find_bin_aperture(
                        focal_plane,
                        ene,
                        spatial_with=dwl[i],
                        center=(wlc[i], spc[i]),
                    )
                    dsp[i] *= size

        new_tab = QTable()
        new_tab["spectral_center"] = wlc
        new_tab["spectral_size"] = dwl
        new_tab["spatial_center"] = spc
        new_tab["spatial_size"] = dsp
        new_tab["aperture_shape"] = shape

        return new_tab



import os
from collections import OrderedDict
from typing import Tuple, Union

import h5py

import exosim.log as log
from exosim.models.signal import CountsPerSecond
from exosim.output import HDF5OutputGroup, SetOutput
from exosim.output.hdf5.utils import load_signal
from exosim.tasks.astrosignal import (
    ApplyAstronomicalSignal,
    FindAstronomicalSignals,
)
from exosim.tasks.subexposures import (
    AddForegrounds,
    ApplyQeMap,
    EstimatePointingJitter,
    InstantaneousReadOut,
    LoadILS,
    LoadQeMap,
    PrepareInstantaneousReadOut,
)
from exosim.utils.ascii_arts import astronomer2
from exosim.utils.iterators import iterate_over_chunks
from exosim.utils.klass_factory import find_and_run_task
from exosim.utils.prepare_recipes import (
    clean_config_files,
    copy_input_files,
    load_options,
)
from exosim.utils.runConfig import RunConfig
from exosim.utils.timed_class import TimedClass


class CreateSubExposures(TimedClass, log.Logger):
    """
    Pipeline to create the focal planes sub-exposures.
    This pipeline loads the configuration file and the produced focal planes to produce the sub-exposures.
    It prepares the pointing jitter first, then it scales it to the instrument focal planes pixels.
    It estimates the detector reading scheme, considering instantaneous readout,
    and it produces the sub-exposures by jittering the focal planes.

    Attributes
    ------------
    input: str
        input file
    mainConfig: dict
        This is parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
    payloadConfig: dict
        payload configuration dictionary extracted from mainConfig
    jitter_spa: :class:`~astropy.units.Quantity`
        pointing jitter in the spatial direction expressed in units of :math:`deg`.
    jitter_spe: :class:`~astropy.units.Quantity`
        pointing jitter in the spectral direction expressed in units of :math:`deg`.


    Examples
    --------

    >>> import exosim.recipes as recipes
    >>> recipes.CreateSubExposures(input_file='./input_file.h5',
    >>>                            output_file = 'se_file.h5',
    >>>                            options_file='main _configuration.xml')


    """

    def __init__(
        self, input_file: str, output_file: str, options_file: Union[str, dict]
    ) -> None:
        """
        Parameters
        ----------
        input_file: str
            input file
        output_file: str
            output file
        options_file: str or dict
            input configuration file
        """

        super().__init__()
        self.graphics(astronomer2)
        RunConfig.stats()
        self.announce("started")
        clean_config_files()
        self.input = input_file
        self.mainConfig, self.payloadConfig = load_options(options_file)
        copy_input_files(os.path.dirname(os.path.abspath(output_file)))

        self.jitter_spa, self.jitter_spe, self.jitter_time = None, None, None
        output = SetOutput(output_file)

        # prepare channel list
        with h5py.File(self.input, "r") as f:
            ch_list = list(f["channels"].keys())
            ch_list.sort()

        with output.use(append=True, cache=True) as out:
            # if jitter is requested, run the jitter task
            use_slicing = False
            if "jitter" in self.mainConfig.keys():
                jitter_instance = find_and_run_task(
                    self.mainConfig["jitter"],
                    "jitter_task",
                    EstimatePointingJitter,
                )
                (
                    self.jitter_spa,
                    self.jitter_spe,
                    self.jitter_time,
                ) = jitter_instance(
                    parameters=self.mainConfig, output_file=out
                )

                if "slicing" in self.mainConfig["jitter"].keys():
                    use_slicing = self.mainConfig["jitter"]["slicing"]

            output_grp = out.create_group("channels")
            instantaneousReadOut = InstantaneousReadOut()
            prepareInstantaneousReadOut = PrepareInstantaneousReadOut()
            addForegrounds = AddForegrounds()

            for ch in ch_list:
                self.announce("computing sub-exposure for {}".format(ch))
                # prepare the output group
                ch_grp = output_grp.create_group(ch)

                # copy the simulation data from the input file to the output file
                sim_grp = self.copy_simulation_data(ch, ch_grp)

                # load the focal plane
                fp, frg_fp, bkg_fp = self.load_focal_plane(ch)

                # determine if only a channel is provided or a list of channels
                if isinstance(self.payloadConfig["channel"], OrderedDict):
                    ch_param = self.payloadConfig["channel"][ch]
                else:
                    ch_param = self.payloadConfig["channel"]

                # prepare the readout task
                (
                    readout_parameters,
                    integration_time,
                ) = prepareInstantaneousReadOut(
                    main_parameters=self.mainConfig,
                    parameters=ch_param,
                    focal_plane=fp,
                    pointing_jitter=(
                        self.jitter_spa,
                        self.jitter_spe,
                        self.jitter_time,
                    ),
                    output_file=ch_grp,
                )

                # run the readout task
                out_se = instantaneousReadOut(
                    parameters=ch_param,
                    focal_plane=fp,
                    readout_parameters=readout_parameters,
                    pointing_jitter=(
                        self.jitter_spa,
                        self.jitter_spe,
                        self.jitter_time,
                    ),
                    slicing=use_slicing,
                    output_file=ch_grp,
                )

                # introducing the astronomical signal
                ils, timeline = None, None

                findAstronomicalSignals = FindAstronomicalSignals()
                applyAstronomicalSignal = ApplyAstronomicalSignal()
                astrosignals = findAstronomicalSignals(
                    sky_parameters=self.mainConfig["sky"],
                )
                for source, signals in astrosignals.items():
                    astro_grp = ch_grp.create_group("Astrosignal")
                    source_grp = astro_grp.create_group(source)
                    # TODO what if different signals are found for the same source with the same name?
                    for signal, params in signals.items():
                        signal_grp = source_grp.create_group(signal)
                        self.info(
                            "{}-{} astronomical signal found.".format(
                                source, signal
                            )
                        )

                        # load the instrument line shape
                        ils_instance = find_and_run_task(
                            ch_param["detector"], "ils_task", LoadILS
                        )
                        ils = (
                            ils_instance(
                                input_file=self.input,
                                ch_param=ch_param,
                                wl_grid=fp.spectral * fp.spectral_units,
                            )
                            if ils is None
                            else ils
                        )

                        # prepare the timeline for the astronomical signal as the middle of each sub-exposure
                        timeline = (
                            (
                                out_se.time * out_se.time_units
                                + integration_time / 2
                            )
                            if timeline is None
                            else timeline
                        )

                        # estimating the signal model
                        signal_task = params["task"]()
                        model_timeline, signal_model = signal_task(
                            timeline=timeline,
                            wl_grid=fp.spectral * fp.spectral_units,
                            ch_parameters=ch_param,
                            source_parameters=params["parsed_parameters"],
                            output=signal_grp,
                        )

                        # load the source flux to weight the signal model
                        with h5py.File(self.input, "r") as f:
                            source_dic = f[f"channels/{ch}/sources/{source}"]
                            flux = source_dic["data"][0, 0]

                        # apply the signal model to the sub-exposures
                        out_se = applyAstronomicalSignal(
                            model=signal_model,
                            subexposures=out_se,
                            focal_plane=fp,
                            ils=ils,
                            timeline=model_timeline,
                            source=params,
                            source_flux=flux,
                            ch_parameters=ch_param,
                            pointing=(
                                self.mainConfig["pointing"]["ra"],
                                self.mainConfig["pointing"]["dec"],
                            ),
                        )

                # adding background stars if any
                add_background_to_se = False
                if "add_background_to_se" in ch_param["detector"].keys():
                    if ch_param["detector"]["add_background_to_se"]:
                        add_background_to_se = True
                else:
                    if bkg_fp:
                        self.warning(
                            "No indication for adding background found. Adding them by default."
                        )
                        add_background_to_se = True

                if bkg_fp and add_background_to_se:
                    bkg_se = instantaneousReadOut(
                        parameters=ch_param,
                        focal_plane=bkg_fp,
                        readout_parameters=readout_parameters,
                        pointing_jitter=(
                            self.jitter_spa,
                            self.jitter_spe,
                            self.jitter_time,
                        ),
                        dataset_name="sub_exposures_bkg",
                        output_file=ch_grp,
                    )

                    for chunk in iterate_over_chunks(
                        out_se.dataset, desc="adding background stars"
                    ):
                        out_se.dataset[chunk] += bkg_se.dataset[chunk]
                        out_se.output.flush()

                # add the foregrounds to the sub-exposures
                try:
                    if ch_param["detector"]["add_foregrounds_to_se"]:
                        out_se = addForegrounds(
                            subexposures=out_se,
                            frg_focal_plane=frg_fp,
                            integration_time=integration_time,
                        )
                except KeyError:
                    self.warning(
                        "No indication for adding foregrounds found. Adding them by default."
                    )
                    out_se = addForegrounds(
                        subexposures=out_se,
                        frg_focal_plane=frg_fp,
                        integration_time=integration_time,
                    )

                # apply the QE map if present
                try:
                    qe_map_instance = find_and_run_task(
                        ch_param["detector"], "qe_map_task", LoadQeMap
                    )
                    qe_map = qe_map_instance(
                        parameters=ch_param,
                        time=fp.time * fp.time_units,
                        output=sim_grp,
                    )
                    applyQEMap = ApplyQeMap()
                    out_se = applyQEMap(subexposures=out_se, qe_map=qe_map)
                except KeyError:
                    self.warning(
                        "No quantum efficiency variation map detected"
                    )
                # store the results in output
                out_se.write()

                self.log_runtime("{} ended in".format(ch), "info")

            self.info(
                "output {} size: {:.3f}".format(output_file, out.getsize())
            )
        self.log_runtime_complete("recipe ended", "info")
        self.announce("ended")

    def load_focal_plane(
        self, ch: str
    ) -> Tuple[CountsPerSecond, CountsPerSecond]:
        """
        It loads the channel focal plane from the input file:

        Parameters
        ----------
        ch: str
            channel name

        Returns
        -------
        :class:`~exosim.models.signal.CountsPerSecond`
            focal plane
        :class:`~exosim.models.signal.CountsPerSecond`
            foreground focal plane
        :class:`~exosim.models.signal.CountsPerSecond`
            bkg focal plane"""
        with h5py.File(self.input, "r") as f:
            file_path = os.path.join("channels", ch)
            fp = load_signal(f[os.path.join(file_path, "focal_plane")])
            frg_fp = load_signal(f[os.path.join(file_path, "frg_focal_plane")])
            try:
                bkg_fp = load_signal(
                    f[os.path.join(file_path, "bkg_focal_plane")]
                )
            except KeyError:
                bkg_fp = None

        self.debug("focal planes loaded from {}".format(ch))

        return fp, frg_fp, bkg_fp

    def load_source(self, ch: str, source: str) -> Tuple[CountsPerSecond]:
        """
        It loads the channel focal plane from the input file:

        Parameters
        ----------
        ch: str
            channel name
        source: str
            source name

        Returns
        -------
        :class:`~exosim.models.signal.CountsPerSecond`
            source
        """
        with h5py.File(self.input, "r") as f:
            file_path = os.path.join("channels", ch)
            path = os.path.join(file_path, "sources")
            source = load_signal(f[os.path.join(path, source)])

        self.debug("{} source loaded from {}".format(source, ch))

        return source

    def copy_simulation_data(
        self, ch: str, ch_grp: HDF5OutputGroup
    ) -> HDF5OutputGroup:
        """
        It copies relevant data from the input file into the output one.
        The copied data are: `info`, `qe_map`, `efficiency` and `responsivity`.

        Parameters
        ----------
        ch: str
            channel name
        ch_grp: :class:`~exosim.output.hdf5.hdf5.HDF5OutputGroup`
            output file

        Returns
        --------
        :class:`~exosim.output.hdf5.hdf5.HDF5OutputGroup`
            output group

        """
        sim_grp = ch_grp.create_group("simulation_data")

        with h5py.File(self.input, "r") as f:
            # copy the original data info
            f.copy("info", sim_grp._entry, name="focal_plane_info")

            # copy simulation data
            file_path = os.path.join("channels", ch)
            for info in ["efficiency", "responsivity"]:
                try:
                    f.copy(os.path.join(file_path, info), sim_grp._entry)
                    self.debug("{} data copied to output".format(info))
                except KeyError:
                    continue
        return sim_grp



from copy import deepcopy

import numpy as np

from exosim.models.signal import Adu
from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks


class AnalogToDigital(Task):
    """
    It converts the :math:`counts` units of sub-exposures into :math:`adu` units of NDRs.
    The conversion is related to the user defined number of bits of the ADC.
    In the number of bits is not defined, 32 is the default value.
    This Task select the smallest ``dtype`` to represent the desired data type:

    + if the input number of bits is smaller than 16, a ``int16`` data type is used,
    + if the input number of bits is smaller than 8, a ``int8`` data type is used.

    Otherwise, a ``int32`` data type is used.

    The user should also specify the ADC gain factor to apply to the focal plane before the conversion.

    Finally, the user should specify the rounding method to use to cast the float into integers.
    The `ADC_round_method` keyword indicates which method the ADC should use to cast the float into integers. Three options are available:

    - `floor` which uses :class:`numpy.floor`;
    - `ceil` which uses :class:`numpy.ceil`;
    - `round` which uses :class:`numpy.round`;

    Default is `floor`.


    Returns
    ---------
    :class:`~exosim.models.signal.Adu`
        sub-exposures converted into NDRs :math:`adu` units
    """

    def __init__(self):
        """
        Parameters
        ----------
        """
        self.add_task_param("subexposures", " ")
        self.add_task_param(
            "parameters",
            " ",
        )
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("converting to digital")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        output = self.get_task_param("output")

        # select the rounding method. Default is 'floor'
        rounders = {"floor": np.floor, "ceil": np.ceil, "round": np.round}
        round_method = "floor"
        if "ADC_round_method" in parameters["detector"].keys():
            round_method = parameters["detector"]["ADC_round_method"]
            if round_method not in ["round", "floor", "ceil"]:
                self.error("round method should be round, floor or ceil")
                raise ValueError("round method should be round, floor or ceil")
        rounder = rounders[round_method]

        if "ADC_num_bit" in parameters["detector"].keys():
            bits_num = parameters["detector"]["ADC_num_bit"]
            if isinstance(bits_num, float):
                if not bits_num.is_integer():
                    self.error("number of ADC bits should be integer")
                    raise TypeError("number of ADC bits should be integer")
            self.debug(
                "ADC set to {} bits.".format(
                    parameters["detector"]["ADC_num_bit"]
                )
            )
        else:
            bits_num = 32
            self.debug("ADC set to 32 bit by dfault.")

        if bits_num > 32:
            self.error("max bits number supported is 32.")
            raise ValueError("max bits number supported is 32.")

        dtype_range = 2**bits_num - 1

        int_type = np.dtype("uint32")
        if bits_num <= 16:
            int_type = np.dtype("uint16")
        if bits_num <= 8:
            int_type = np.dtype("uint8")

        gain_factor = 1.0
        if "ADC_gain" in parameters["detector"]:
            if isinstance(parameters["detector"]["ADC_gain"], str):
                gain_factor = parameters["detector"]["ADC_gain"].lower()
            else:
                gain_factor = parameters["detector"]["ADC_gain"]

        offset = "auto"
        if "ADC_offset" in parameters["detector"]:
            if isinstance(parameters["detector"]["ADC_offset"], str):
                offset = parameters["detector"]["ADC_offset"].lower()
            else:
                offset = parameters["detector"]["ADC_offset"]

        if offset == "auto":
            offset = np.min(subexposures.dataset)

        if gain_factor == "auto":
            max_ = np.max(subexposures.dataset)

            gain_factor = dtype_range / (max_ - offset)
            gain_factor = float(gain_factor)

        self.info("ADC gain:{}".format(gain_factor))
        self.info("ADC offset:{}".format(offset))
        self.info("ADC range: {}".format(dtype_range))

        ndrs = Adu(
            spectral=subexposures.spectral,
            time=subexposures.time,
            data=None,
            spatial=subexposures.spatial,
            shape=subexposures.shape,
            cached=True,
            output=output,
            dataset_name="ADC_ndrs",
            metadata=subexposures.metadata,
            output_path=None,
            dtype=int_type,
        )

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="converting to digital"
        ):
            data = (
                deepcopy(subexposures.dataset[chunk]) - offset
            ) * gain_factor
            data[data < 0] = 0  # applying the offset as trigger
            ndrs.dataset[chunk] = rounder(data).astype(int_type)
            ndrs.output.flush()

        ndrs.metadata["ADC"] = {"gain": gain_factor, "offset": offset}

        self.set_output(ndrs)



from .addForegrounds import AddForegrounds
from .applyQeMap import ApplyQeMap
from .computeReadingScheme import ComputeReadingScheme
from .estimateChJitter import EstimateChJitter
from .estimatePointingJitter import EstimatePointingJitter
from .instantaneousReadOut import InstantaneousReadOut
from .loadILS import LoadILS
from .loadQeMap import LoadQeMap
from .loadQeMapNumpy import LoadQeMapNumpy
from .prepareInstantaneousReadOut import PrepareInstantaneousReadOut



import astropy.units as u
import numpy as np


def check_units(input_data, desired_units, calling_class=None, force=False):
    """
    It checks the units of the inputs and returns the quantity rescaled to the desired units.

    Parameters
    ----------
    input_data: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
        input class to rescale.
    desired_units: :class:`~astropy.units.Quantity`
        desired unit for the input qquantity.
    calling_class: :class:`~exosim.log.logger.Logger` (optional)
        calling class. This is needed to print the eventual debug message inside the calling class.
    force: bool (optional)
        if True, if the input data has no units, it assumes is expressed in the desired units. Default is False.

    Returns
    -------
    :class:`~astropy.units.Quantity`
        scaled input quantity.

    Raises
    ------
    UnitConversionError
        if it cannot convert the original units into the desired ones
    """

    def _has_no_unit(_input_data, _desired_units, _force):
        """If input has no unit, it assigns the desired one, if force is True,
        or just convert the input data to a :class:`~astropy.units.Quantity` with no unit.
        """
        if calling_class:
            calling_class.debug("input data has no unit")
        if _force:
            if calling_class:
                calling_class.debug("forcing {} units".format(_desired_units))
            _input_data = np.array(_input_data) * u.Unit(_desired_units)
        else:
            if calling_class:
                calling_class.debug("forcing no units")
            _input_data = np.array(_input_data) * u.Unit("")
        return _input_data

    def _has_different_unit(_input_data, _desired_units):
        """If input has different unit, it converts the input data into the desired one."""
        try:
            output_data = _input_data.to(u.Unit(_desired_units))
            if calling_class:
                calling_class.debug(
                    "converted {} to {}".format(
                        _input_data.unit, _desired_units
                    )
                )
                calling_class.debug("converted data: {}".format(output_data))
            return output_data

        except u.UnitConversionError:
            if _input_data.unit == u.s and _desired_units == u.Hz:
                return check_units(1 / _input_data, _desired_units)
            elif _input_data.unit == u.Hz and _desired_units == u.s:
                return check_units(1 / _input_data, _desired_units)
            else:
                msg = "impossible to convert {} into {}".format(
                    _input_data.unit, _desired_units
                )
                raise u.UnitConversionError(msg)

    if desired_units is None:
        desired_units = ""

    if not hasattr(input_data, "unit"):
        input_data = _has_no_unit(input_data, desired_units, force)
    elif input_data.unit is None:
        input_data = _has_no_unit(input_data, desired_units, force)
    else:
        input_unit = input_data.unit
        if input_unit == 1 / u.Hz:
            input_unit = u.s
        if input_unit == 1 / u.s:
            input_unit = u.Hz
        if not input_unit:
            input_unit = u.Unit("")
        input_data = np.array(input_data) * u.Unit(input_unit)

    if calling_class:
        calling_class.debug("input data: {}".format(input_data))

    if input_data.unit != u.Unit(desired_units):
        return _has_different_unit(input_data, desired_units)

    else:
        return input_data


def find_key(input_class_keys, key_list, calling_class=None):
    """
    Finds which key from key_list is contained in input_class_keys, ignoring case.

    Parameters
    ----------
    input_class_keys : list
        List of the input class keys.
    key_list : list
        List of the desired keys.
    calling_class : :class:`~exosim.log.logger.Logger`, optional
        Calling class. This is needed to print the eventual debug message inside the calling class.

    Returns
    -------
    str
        Found key.

    Raises
    -------
    KeyError
        If no matching key is found.
    """
    # Create a set of lower-case keys from the input class for efficient lookup
    lower_input_keys = {key.lower() for key in input_class_keys}

    try:
        # Find the first key from key_list that exists in lower_input_keys (case-insensitive)
        key = [k for k in key_list if k.lower() in lower_input_keys][0]
        return key
    except IndexError:
        msg = "no matching key found"
        if calling_class:
            calling_class.error(msg)
        raise KeyError(msg)


def look_for_key(input_dict, key, value, foo=False):
    """
    Returns ``True`` if a certain key is in the dictionary and has a certain value.

    Parameters
    ----------
    input_dict: dict
        input dictionary
    key: str
        key to search
    value
        key content to check

    Returns
    -------
    bool

    """
    for k, val in input_dict.items():
        if k == key and val == value:
            foo = True
            break
        else:
            if isinstance(input_dict[k], dict):
                foo = look_for_key(input_dict[k], key, value, foo)
    return foo



import numpy as np

import exosim.models.signal as signal
import exosim.output as output
import exosim.tasks.foregrounds as foregrounds
from exosim.tasks.task import Task


class ParseZodi(Task):
    """
    This tasks parses the zodiacal foreground.

    Returns
    -------
    dict
        dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
        representing the radiance and efficiency of the optical element.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        self.add_task_param("parameters", "optical element parameters dict")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file", None)

    def execute(self):
        parameters = self.get_task_param("parameters")
        self.debug("parsing: {}".format(parameters["value"]))

        wl = self.get_task_param("wavelength")
        tt = self.get_task_param("time")

        estimateZodi = foregrounds.EstimateZodi()
        # try to use zodiacal factor
        try:
            radiance = estimateZodi(
                wavelength=wl, zodiacal_factor=parameters["zodiacal_factor"]
            )
        # else, it try to use the map
        except KeyError:
            if "zodi_map" in parameters.keys():
                radiance = estimateZodi(
                    wavelength=wl,
                    coordinates=parameters["coordinates"],
                    zodi_map=parameters["zodi_map"],
                )
            else:
                radiance = estimateZodi(
                    wavelength=wl, coordinates=parameters["coordinates"]
                )
        efficiency = signal.Dimensionless(spectral=wl, data=np.ones(wl.size))
        efficiency.temporal_rebin(tt)

        output_file = self.get_task_param("output")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                og_ = output_file.create_group(parameters["value"])
                radiance.write(og_, "radiance")
                efficiency.write(og_, "efficiency")

        out = {"radiance": radiance, "efficiency": efficiency}

        self.set_output(out)



import logging
import os

import exosim.plots as plots
from exosim import __version__ as version

logger = logging.getLogger("exosim")


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser(description="ExoSim {}".format(version))
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        type=str,
        required=True,
        help="Input file to pass",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        required=True,
        default=None,
        help="Output file",
    )
    parser.add_argument(
        "-f",
        "--focal_plane",
        dest="focal",
        default=False,
        required=False,
        help="run to plot the focal plane ",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--radiometric",
        dest="radiometric",
        default=False,
        required=False,
        help="run to plot the radiometric table",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--subexposures",
        dest="subexposures",
        default=False,
        required=False,
        help="run to plot the sub-exposures",
        action="store_true",
    )
    parser.add_argument(
        "-n",
        "--ndrs",
        dest="ndrs",
        default=False,
        required=False,
        help="run to plot the NDRs plotter",
        action="store_true",
    )
    parser.add_argument(
        "-t",
        "--time_step",
        dest="time_step",
        type=int,
        default=0,
        required=False,
        help="focal plane time step to plot ",
    )

    parser.add_argument(
        "--plot-scale",
        dest="plot_scale",
        type=str,
        default="linear",
        required=False,
        help="plot scale. Can be 'linear' or 'dB'. Default is 'linear' ",
    )
    args = parser.parse_args()

    if args.focal:
        time = args.time_step
        scale = args.plot_scale
        focal_plane_plotter = plots.FocalPlanePlotter(input=args.input)
        focal_plane_plotter.plot_focal_plane(time_step=time, scale=scale)
        fig_name = os.path.join(args.output, "focal_plane_{}.png".format(time))
        focal_plane_plotter.save_fig(fig_name)

        focal_plane_plotter.plot_efficiency()
        fig_name = os.path.join(args.output, "efficiency.png")
        focal_plane_plotter.save_fig(fig_name)

    if args.radiometric:
        radiometric_plotter = plots.RadiometricPlotter(input=args.input)
        radiometric_plotter.plot_table()
        fig_name = os.path.join(args.output, "radiometric.png")
        radiometric_plotter.save_fig(fig_name)

        radiometric_plotter.plot_apertures()
        fig_name = os.path.join(args.output, "apertures.png")
        radiometric_plotter.save_fig(fig_name)

    if args.subexposures:
        sub_exposures_plotter = plots.SubExposuresPlotter(input=args.input)
        sub_exposures_plotter.plot(args.output)

    if args.ndrs:
        ndrs_plotter = plots.NDRsPlotter(input=args.input)
        ndrs_plotter.plot(args.output)



from copy import deepcopy

import numpy as np

from exosim.output import Output
from exosim.tasks.task import Task
from exosim.utils import RunConfig
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks


class AddNormalReadNoise(Task):
    """
    This Task simulates the read noise as a normal distribution which parameters can be defined in the configuration file.

    If it is not explicitly stated, the distribution mean is set to 0.
    A different realisation of the same distribution is added to each pixel of each sub-exposure.
    If an output group is provided, it saves all the random seeds used.

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding read noise")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        output = self.get_task_param("output")

        read_noise_sigma = parameters["detector"]["read_noise_sigma"].astype(
            np.float64
        )
        read_noise_sigma = check_units(read_noise_sigma, "ct").value

        random_seeds = []

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding read noise"
        ):
            data = deepcopy(subexposures.dataset[chunk])
            subexposures.dataset[chunk] = (
                data
                + RunConfig.random_generator.normal(
                    0, read_noise_sigma, data.shape
                ).astype(np.float64)
            )
            subexposures.output.flush()
            random_seeds.append(RunConfig.random_seed)

        if output:
            if issubclass(output.__class__, Output):
                out_grp = output.create_group("read noise")
                out_grp.write_list("random_seed", random_seeds)
                out_grp.write_array(
                    "chunks_index", np.arange(len(random_seeds))
                )



from .createCustomSource import CreateCustomSource
from .createPlanckStar import CreatePlanckStar
from .loadCustom import LoadCustom
from .loadPhoenix import LoadPhoenix
from .prepareSed import PrepareSed



import astropy.units as u
import numpy as np

import exosim.output as output
from exosim.tasks.task import Task


class CreateIntrapixelResponseFunction(Task):
    """
    It creates the intrapixel response function to be used in the convolution with the focal plane array.


    Returns
    -------
    kernel	: 2D array
            the kernel image
    kernel_delta  : scalar
                    the kernel sampling interval in microns


    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        output: :class:`~exosim.output.output.Output` (optional)
           output file
        """

        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("creating pixel respone function")
        parameters = self.get_task_param("parameters")

        kernel, kernel_delta = self.model(parameters)

        output_file = self.get_task_param("output")
        if output_file and issubclass(output_file.__class__, output.Output):
            output_group = output_file.create_group(
                "intrapixel_response_function"
            )
            output_group.write_array(
                "kernel",
                kernel,
            )
            output_group.write_quantity(
                "kernel_delta",
                kernel_delta,
            )

        self.set_output([kernel, kernel_delta])

    def model(self, parameters):
        """
        This model creates the intrapixel response function to be used in the convolution with the focal plane array.
        Thsi intrapixel response function is compatible with the Scipy convolution functions.

        Estimate the detector pixel response function with the prescription of
        Barron et al., PASP, 119, 466-475 (2007).

        Parameters
        ----------
        oversampling: int
                number of samples in each resolving element. The
                final shape of the response function would be shape*osf
        delta_pix: :class:`astropy.units.Quantity`
                Phisical size of the detector pixel in microns
        diffusion_length: :class:`astropy.units.Quantity`
                diffusion length in microns
        intra_pix_distance: :class:`astropy.units.Quantity`
                distance between two adjacent detector pixels
                in microns

        Returns
        -------
        kernel	: 2D array
                the kernel image
        kernel_delta  : :class:`astropy.units.Quantity`
                        the kernel sampling interval in microns
        """
        if "oversampling" in parameters["detector"].keys():
            osf = parameters["detector"][
                "oversampling"
            ]  # activate for old convolution
        else:
            self.warning(
                "oversampling is not defined in the detector parameters. Oversampling factor is set to 1."
            )
            osf = 1  # activate for old convolution

        delta = parameters["detector"]["delta_pix"]
        if "diffusion_length" in parameters["detector"].keys():
            lx = parameters["detector"]["diffusion_length"]
        else:
            lx = 0.0 * u.um
        if "intra_pix_distance" in parameters["detector"].keys():
            ipd = parameters["detector"]["intra_pix_distance"]
        else:
            ipd = 0 * u.um
        lx += 1e-20 * u.um  # to avoid problems if user pass lx=0
        lx = lx.to(delta.unit)

        kernel = np.zeros((osf, osf))
        self.debug("kernel size: {}".format(kernel.shape))
        # prepare the kernel stamp grid
        kernel_delta = delta / osf
        scale = np.arange(0, 2 * osf / 2) - osf / 2 + 0.5
        yy = scale * kernel_delta
        xx = scale * kernel_delta

        xx, yy = np.meshgrid(xx, yy)

        # compute kernel stamp
        kernel_stamp = np.arctan(
            np.tanh((0.5 * (0.5 * delta - xx) / lx).value)
        ) - np.arctan(np.tanh((0.5 * (-0.5 * delta - xx) / lx).value))

        kernel_stamp *= np.arctan(
            np.tanh((0.5 * (0.5 * delta - yy) / lx).value)
        ) - np.arctan(np.tanh((0.5 * (-0.5 * delta - yy) / lx).value))

        # deal with border
        i_mask_xx = np.where(np.abs(xx) > 0.5 * (delta - ipd))
        i_mask_yy = np.where(np.abs(yy) > 0.5 * (delta - ipd))
        # set the unused area of kernel stamp to zero
        kernel_stamp[i_mask_yy] = 0.0
        kernel_stamp[i_mask_xx] = 0.0

        # # deal with fractional pixels
        # border = kernel_delta - ipd
        # pix, fract = divmod(border.value,1)

        # if fract != 0:
        #     idx = np.where(kernel_stamp!=0)
        #     kernel_stamp[:, idx[0][0]] *= 1-fract
        #     kernel_stamp[idx[1][0], :] *= 1-fract
        #     kernel_stamp[:, idx[0][-1]] *= 1-fract
        #     kernel_stamp[idx[1][-1], :] *= 1-fract
        # Normalise the kernel such that the pixel has QE=1
        kernel_stamp /= kernel_stamp.sum()
        kernel_stamp *= osf * osf

        return kernel_stamp, kernel_delta

        # Author note: We finally moved to a more general approach using scipy fftconvolve.
        # In thi vision we can simulate a kernel of a single pixel.
        # The problem with previuos approach is that if the kernel was not zeri padded as the PSF the results will have aliasing effects.
        # 2022-12-01 L.V.M

        # Author note: previous version was computing the pixel irf shape on a kernel with the size of the PSF shape.
        # This was numerical intensive and not necessary. The new version is much faster and more accurate.
        # The kernel is now a square with the size of 4 pixels. The kernel is then rolled to use for the fast fft convolve.
        # Previous version can be tracked in the git history.
        # 2022-11-30 L.V.M

        # Author note: previous version was computing the pixel irf shape on the full kernel and then setting to 0 all other pixels.
        # That method resulted in a very slow computation of arctan and huge memory requiring for high osf.
        # This "kernel stamp" solution return the same kernel (with relative differences under 1e-8 - tested on 2022/05/19)
        # but very lower computing time.
        # TEST: using osf 12 psf =(64,64): previous method took 4.2s, new method takes 6ms with)
        # This method removes the memory problem: not much of RAM used for this computation
        # tested on jupyter notebook on 2022/05/19. L.V.M.



import astropy.units as u
import numpy as np

from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks, searchsorted


class ApplyQeMap(Task):
    """
    It applies the quantum efficiency variation map to the sub exposures.

    Returns
    --------
    :class:`~exosim.models.signal.Counts`
        sub-exposure cached signal class
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposure cached signal class
        qe_map: :class:`~exosim.models.signal.Signal`
            channel responsivity variation map
        """

        self.add_task_param("subexposures", "sub-exposure cached signal clas")
        self.add_task_param(
            "qe_map",
            "channel responsivity variation map",
        )

    def execute(self):
        subexposures = self.get_task_param("subexposures")
        qe_map = self.get_task_param("qe_map")

        subexposures_time = (subexposures.time * subexposures.time_units).to(
            u.hr
        )
        qe_map_time = qe_map.time
        if not isinstance(qe_map_time, u.Quantity):
            qe_map_time = qe_map.time * qe_map.time_units
        qe_map_time.to(u.hr)
        qe = qe_map.data.astype(np.float64)
        index = searchsorted(qe_map_time, subexposures_time)

        self.debug("time indexes: {}".format(index))
        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="applying qe variation map"
        ):
            subexposures.dataset[chunk] = self.apply_qe(
                subexposures.dataset[chunk], qe, index[chunk[0]]
            )
            subexposures.output.flush()
        subexposures.metadata["qe_variation_map_indexes"] = index
        self.set_output(subexposures)

    @staticmethod
    def apply_qe(se, qe, index):
        for t in range(se.shape[0]):
            se[t] *= qe[index[t]]
        return se



import random

import numpy as np
from numba import get_num_threads, set_num_threads

from exosim.log import Logger

total_cpus = get_num_threads()


# Singleton Meta-Class method
# https://refactoring.guru/design-patterns/singleton/python/example


class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class RunConfigInit(Logger, metaclass=SingletonMeta):
    """
    Class used to propagate values through the code.
    """

    _n_job = 1
    chunk_size = 2
    _random_seed = None
    config_file_list = []

    def __init__(self):
        self.set_log_name()

    @property
    def random_seed(self):
        return self._random_seed

    @random_seed.setter
    def random_seed(self, seed):
        np.random.seed(seed)
        random.seed(seed)
        self._random_seed = seed

    @random_seed.deleter
    def random_seed(self):
        del self._random_seed

    @property
    def n_job(self):
        return self._n_job

    @n_job.setter
    def n_job(self, value):
        value = value if value > 0 else total_cpus + value
        set_num_threads(value)
        self._n_job = value

    @property
    def random_generator(self):
        """Returns a random generator with the current random seed updated by one if the random seed is set"""
        if self.random_seed is not None:
            self.random_seed += 1
        return np.random.default_rng(seed=self.random_seed)

    def __dict__(self):
        out = {}
        for a in ["n_job", "chunk_size", "random_seed", "config_file_list"]:
            out[a] = self.__getattribute__(a)
        return out

    def stats(self, log=True):
        """
        It returns a dictionary with the updated run configurations.

        Parameters
        ----------
        log: bool (optional)
            if True, it prints the run configuration stats. Default is True.

        Returns
        -------

        """
        out_dict = {
            "number of available cpus": total_cpus,
            "number of used cpus": self.n_job,
            "random seed": self.random_seed,
            "chunk size (Mb)": self.chunk_size,
        }
        if log:
            self.info("-- RUN CONFIGURATION ---------")
            for key, val in out_dict.items():
                self.info("{}: {}".format(key, val))
            self.info("------------------------------")

        return out_dict


RunConfig = RunConfigInit()
RunConfig.random_seed = random.randrange(0, 2**32 - 1)
